# Copyright 2021 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
import webob.exc

from glance.api import policy
from glance.common import exception
from glance.i18n import _

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


# TODO(danms): Remove this once secure RBAC is fully implemented and
# used instead of legacy policy checks.
def check_is_image_mutable(context, image):
    """Replicate the DB-layer admin-or-owner check for the API.

    Much of the API code depends on hard-coded admin-or-owner
    enforcement in the DB or authorization layer, as the policy layer
    is largely a no-op by default. During blueprint policy-refactor,
    we are trying to remove as much of that as possible, but in
    certain places we need to do that (if secure_rbac is not
    enabled). This transitional helper provides a way to do that
    enforcement where necessary.

    :param context: A RequestContext
    :param image: An ImageProxy
    :raises: exception.Forbidden if the context is not the owner or an admin
    """
    # Is admin == image mutable
    if context.is_admin:
        return

    # No owner == image not mutable
    # Image only mutable by its owner
    if (image.owner is None or context.owner is None or
            image.owner != context.owner):
        raise exception.Forbidden(_('You do not own this image'))


def check_admin_or_same_owner(context, properties):
    """Check that legacy behavior on create with owner is preserved.

    Legacy behavior requires a static check that owner is not
    inconsistent with the context, unless the caller is an
    admin. Enforce that here, if needed.

    :param context: A RequestContext
    :param properties: The properties being used to create the image, which may
                       contain an owner
    :raises: exception.Forbidden if the context is not an admin and owner is
             set to something other than the context's project
    """
    if context.is_admin:
        return

    if context.project_id != properties.get('owner', context.project_id):
        msg = _("You are not permitted to create images "
                "owned by '%s'.")
        raise exception.Forbidden(msg % properties['owner'])


class APIPolicyBase(object):
    def __init__(self, context, target=None, enforcer=None):
        self._context = context
        self._target = target or {}
        self.enforcer = enforcer or policy.Enforcer()

    def _enforce(self, rule_name):
        try:
            self.enforcer.enforce(self._context, rule_name, self._target)
        except exception.Forbidden as e:
            raise webob.exc.HTTPForbidden(explanation=str(e))

    def check(self, name, *args):
        """Perform a soft check of a named policy.

        This is used when you need to check if a policy is allowed for the
        given resource, without needing to catch an exception. If the policy
        check requires args, those are accepted here as well.

        :param name: Policy name to check
        :returns: bool indicating if the policy is allowed.
        """
        try:
            getattr(self, name)(*args)
            return True
        except webob.exc.HTTPForbidden:
            return False


class CacheImageAPIPolicy(APIPolicyBase):
    def __init__(self, context, image=None, policy_str=None,
                 target=None, enforcer=None):
        self._context = context
        target = {}
        self._image = image
        if self._image:
            target = policy.ImageTarget(self._image)

        self._target = target
        self.enforcer = enforcer or policy.Enforcer()
        self.policy_str = policy_str
        super(CacheImageAPIPolicy, self).__init__(context, target, enforcer)

    def manage_image_cache(self):
        self._enforce(self.policy_str)


class DiscoveryAPIPolicy(APIPolicyBase):
    def __init__(self, context, target=None, enforcer=None):
        self._context = context
        self._target = target or {}
        self.enforcer = enforcer or policy.Enforcer()
        super(DiscoveryAPIPolicy, self).__init__(context, target, enforcer)

    def stores_info_detail(self):
        self._enforce('stores_info_detail')


class ImageAPIPolicy(APIPolicyBase):
    def __init__(self, context, image, enforcer=None):
        """Image API policy module.

        :param context: The RequestContext
        :param image: The ImageProxy object in question, or a dict of image
                      properties if no image is yet created or needed for
                      authorization context.
        :param enforcer: The policy.Enforcer object to use for enforcement
                         operations. If not provided (or None), the default
                         enforcer will be selected.
        """
        self._image = image
        if not self.is_created:
            # NOTE(danms): If we are being called with a dict of image
            # properties then we are testing policies that involve
            # creating an image or other image-related resources but
            # without a specific image for context. The target is a
            # dict of proposed image properties, similar to the
            # dict-like interface the ImageTarget provides over
            # a real Image object, with specific keys.
            target = {'project_id': image.get('owner', context.project_id),
                      'owner': image.get('owner', context.project_id),
                      'visibility': image.get('visibility', 'private')}
        else:
            target = policy.ImageTarget(image)
        super(ImageAPIPolicy, self).__init__(context, target, enforcer)

    @property
    def is_created(self):
        """Signal whether the image actually exists or not.

        False if the image is only being proposed by a create operation,
        True if it has already been created.
        """
        return not isinstance(self._image, dict)

    def _enforce(self, rule_name):
        """Translate Forbidden->NotFound for images."""
        try:
            super(ImageAPIPolicy, self)._enforce(rule_name)
        except webob.exc.HTTPForbidden:
            # If we are checking image policy before creating an
            # image, or without a specific image for context, then we
            # do not need to potentially hide the presence of anything
            # based on visibility, so re-raise immediately.
            if not self.is_created:
                raise

            # If we are checking get_image, then Forbidden means the
            # user cannot see this image, so raise NotFound. If we are
            # checking anything else and get Forbidden, then raise
            # NotFound in that case as well to avoid exposing images
            # the user can not see, while preserving the Forbidden
            # behavior for the ones they can see.
            if rule_name == 'get_image' or not self.check('get_image'):
                raise webob.exc.HTTPNotFound()
            raise

    def check(self, name, *args):
        try:
            return super(ImageAPIPolicy, self).check(name, *args)
        except webob.exc.HTTPNotFound:
            # NOTE(danms): Since our _enforce can raise NotFound, that
            # too means a False check response.
            return False

    def _enforce_visibility(self, visibility):
        # NOTE(danms): Use the existing enforcement routine for now,
        # which shows that we're enforcing the same behavior. In the
        # future, that should probably be moved here.
        try:
            policy._enforce_image_visibility(self.enforcer, self._context,
                                             visibility, self._target)
        except exception.Forbidden as e:
            raise webob.exc.HTTPForbidden(explanation=str(e))

    def update_property(self, name, value=None):
        if name == 'visibility':
            # NOTE(danms): Visibility changes have their own policy,
            # so check that first, followed by the general
            # modify_image policy below.
            self._enforce_visibility(value)
        self.modify_image()

    def update_locations(self):
        self._enforce('set_image_location')

    def delete_locations(self):
        self._enforce('delete_image_location')
        # TODO(danms): Remove this legacy fallback when secure RBAC
        # replaces the legacy policy.
        if not (CONF.oslo_policy.enforce_new_defaults or
                CONF.oslo_policy.enforce_scope):
            check_is_image_mutable(self._context, self._image)

    def get_image_location(self):
        self._enforce('get_image_location')

    def add_image(self):
        try:
            self._enforce('add_image')
        except webob.exc.HTTPForbidden:
            # NOTE(danms): If we fail add_image because the owner is
            # different, alter the message to be informative and
            # in-line with the current message users have been getting
            # in the past.
            if self._target['owner'] != self._context.project_id:
                msg = _("You are not permitted to create images "
                        "owned by '%s'" % self._target['owner'])
                raise webob.exc.HTTPForbidden(msg)
            else:
                raise
        if 'visibility' in self._target:
            self._enforce_visibility(self._target['visibility'])
        if not (CONF.oslo_policy.enforce_new_defaults or
                CONF.oslo_policy.enforce_scope):
            check_admin_or_same_owner(self._context, self._target)

    def get_image(self):
        self._enforce('get_image')

    def get_images(self):
        self._enforce('get_images')

    def delete_image(self):
        self._enforce('delete_image')
        # TODO(danms): Remove this legacy fallback when secure RBAC
        # replaces the legacy policy.
        if not (CONF.oslo_policy.enforce_new_defaults or
                CONF.oslo_policy.enforce_scope):
            check_is_image_mutable(self._context, self._image)

    def upload_image(self):
        self._enforce('upload_image')
        # TODO(danms): Remove this legacy fallback when secure RBAC
        # replaces the legacy policy.
        if not (CONF.oslo_policy.enforce_new_defaults or
                CONF.oslo_policy.enforce_scope):
            check_is_image_mutable(self._context, self._image)

    def download_image(self):
        self._enforce('download_image')

    def modify_image(self):
        self._enforce('modify_image')
        # TODO(danms): Remove this legacy fallback when secure RBAC
        # replaces the legacy policy.
        if not (CONF.oslo_policy.enforce_new_defaults or
                CONF.oslo_policy.enforce_scope):
            check_is_image_mutable(self._context, self._image)

    def deactivate_image(self):
        self._enforce('deactivate')
        # TODO(danms): Remove this legacy fallback when secure RBAC
        # replaces the legacy policy.
        if not (CONF.oslo_policy.enforce_new_defaults or
                CONF.oslo_policy.enforce_scope):
            check_is_image_mutable(self._context, self._image)

    def reactivate_image(self):
        self._enforce('reactivate')
        # TODO(danms): Remove this legacy fallback when secure RBAC
        # replaces the legacy policy.
        if not (CONF.oslo_policy.enforce_new_defaults or
                CONF.oslo_policy.enforce_scope):
            check_is_image_mutable(self._context, self._image)

    def copy_image(self):
        self._enforce('copy_image')


class MetadefAPIPolicy(APIPolicyBase):
    def __init__(self, context, md_resource=None, target=None, enforcer=None):
        self._context = context
        self._md_resource = md_resource
        if not target:
            self._target = self._build_target()
        else:
            self._target = target
        self.enforcer = enforcer or policy.Enforcer()
        super(MetadefAPIPolicy, self).__init__(context, target=self._target,
                                               enforcer=self.enforcer)

    def _build_target(self):
        target = {
            "project_id": self._context.project_id
        }
        if self._md_resource:
            target['project_id'] = self._md_resource.owner
            target['visibility'] = self._md_resource.visibility

        return target

    def _enforce(self, rule_name):
        """Translate Forbidden->NotFound for images."""
        try:
            super(MetadefAPIPolicy, self)._enforce(rule_name)
        except webob.exc.HTTPForbidden:
            # If we are checking get_metadef_namespace, then Forbidden means
            # the user cannot see this namespace, so raise NotFound. If we are
            # checking anything else and get Forbidden, then raise
            # NotFound in that case as well to avoid exposing namespaces
            # the user can not see, while preserving the Forbidden
            # behavior for the ones they can see.
            if rule_name == 'get_metadef_namespace' or not self.check(
                    'get_metadef_namespace'):
                raise webob.exc.HTTPNotFound()
            raise

    def check(self, name, *args):
        try:
            return super(MetadefAPIPolicy, self).check(name, *args)
        except webob.exc.HTTPNotFound:
            # NOTE(danms): Since our _enforce can raise NotFound, that
            # too means a False check response.
            return False

    def get_metadef_namespace(self):
        self._enforce('get_metadef_namespace')

    def get_metadef_namespaces(self):
        self._enforce('get_metadef_namespaces')

    def add_metadef_namespace(self):
        self._enforce('add_metadef_namespace')

    def modify_metadef_namespace(self):
        self._enforce('modify_metadef_namespace')

    def delete_metadef_namespace(self):
        self._enforce('delete_metadef_namespace')

    def get_metadef_objects(self):
        self._enforce('get_metadef_objects')

    def add_metadef_object(self):
        self._enforce('add_metadef_object')

    def get_metadef_object(self):
        self._enforce('get_metadef_object')

    def modify_metadef_object(self):
        self._enforce('modify_metadef_object')

    def delete_metadef_object(self):
        self._enforce('delete_metadef_object')

    def add_metadef_tag(self):
        self._enforce('add_metadef_tag')

    def get_metadef_tags(self):
        self._enforce('get_metadef_tags')

    def add_metadef_tags(self):
        self._enforce('add_metadef_tags')

    def get_metadef_tag(self):
        self._enforce('get_metadef_tag')

    def modify_metadef_tag(self):
        self._enforce('modify_metadef_tag')

    def delete_metadef_tag(self):
        self._enforce('delete_metadef_tag')

    def delete_metadef_tags(self):
        self._enforce('delete_metadef_tags')

    def add_metadef_property(self):
        self._enforce('add_metadef_property')

    def get_metadef_properties(self):
        self._enforce('get_metadef_properties')

    def remove_metadef_property(self):
        self._enforce('remove_metadef_property')

    def get_metadef_property(self):
        self._enforce('get_metadef_property')

    def modify_metadef_property(self):
        self._enforce('modify_metadef_property')

    def add_metadef_resource_type_association(self):
        self._enforce('add_metadef_resource_type_association')

    def list_metadef_resource_types(self):
        self._enforce('list_metadef_resource_types')

    def get_metadef_resource_type(self):
        self._enforce('get_metadef_resource_type')

    def remove_metadef_resource_type_association(self):
        self._enforce('remove_metadef_resource_type_association')


class MemberAPIPolicy(APIPolicyBase):
    def __init__(self, context, image, target=None, enforcer=None):
        self._context = context
        self._image = image
        if not target:
            self._target = self._build_target()

        self.enforcer = enforcer or policy.Enforcer()
        super(MemberAPIPolicy, self).__init__(context, target=self._target,
                                              enforcer=self.enforcer)

    def _build_target(self):
        target = {
            "project_id": self._context.project_id
        }
        if self._image:
            target = policy.ImageTarget(self._image)

        return target

    def _enforce(self, rule_name):
        ImageAPIPolicy(self._context, self._image,
                       enforcer=self.enforcer).get_image()
        super(MemberAPIPolicy, self)._enforce(rule_name)

    def get_members(self):
        self._enforce("get_members")

    def get_member(self):
        self._enforce("get_member")

    def delete_member(self):
        self._enforce("delete_member")

    def modify_member(self):
        self._enforce("modify_member")

    def add_member(self):
        self._enforce("add_member")


class TasksAPIPolicy(APIPolicyBase):
    def __init__(self, context, target=None, enforcer=None):
        self._context = context
        self._target = target or {}
        self.enforcer = enforcer or policy.Enforcer()
        super(TasksAPIPolicy, self).__init__(context, target=self._target,
                                             enforcer=self.enforcer)

    def tasks_api_access(self):
        self._enforce('tasks_api_access')



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from stevedore import named


CONF = cfg.CONF


def get_import_plugins(**kwargs):
    task_list = CONF.image_import_opts.image_import_plugins
    extensions = named.NamedExtensionManager('glance.image_import.plugins',
                                             names=task_list,
                                             name_order=True,
                                             invoke_on_load=True,
                                             invoke_kwds=kwargs)
    for extension in extensions.extensions:
        yield extension.obj



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_db import exception as db_exc
from oslo_log import log as logging
from sqlalchemy import func
import sqlalchemy.orm as sa_orm

from glance.common import exception as exc
from glance.db.sqlalchemy.metadef_api import namespace as namespace_api
import glance.db.sqlalchemy.metadef_api.utils as metadef_utils
from glance.db.sqlalchemy import models_metadef as models
from glance.i18n import _

LOG = logging.getLogger(__name__)


def _get(context, object_id, session):
    try:
        query = session.query(models.MetadefObject).filter_by(id=object_id)
        metadef_object = query.one()
    except sa_orm.exc.NoResultFound:
        msg = (_("Metadata definition object not found for id=%s")
               % object_id)
        LOG.warning(msg)
        raise exc.MetadefObjectNotFound(msg)

    return metadef_object


def _get_by_name(context, namespace_name, name, session):
    namespace = namespace_api.get(context, namespace_name, session)
    try:
        query = session.query(models.MetadefObject).filter_by(
            name=name, namespace_id=namespace['id'])
        metadef_object = query.one()
    except sa_orm.exc.NoResultFound:
        LOG.debug("The metadata definition object with name=%(name)s"
                  " was not found in namespace=%(namespace_name)s.",
                  {'name': name, 'namespace_name': namespace_name})
        raise exc.MetadefObjectNotFound(object_name=name,
                                        namespace_name=namespace_name)

    return metadef_object


def get_all(context, namespace_name, session):
    namespace = namespace_api.get(context, namespace_name, session)
    query = session.query(models.MetadefObject).filter_by(
        namespace_id=namespace['id'])
    md_objects = query.all()

    md_objects_list = []
    for obj in md_objects:
        md_objects_list.append(obj.to_dict())
    return md_objects_list


def create(context, namespace_name, values, session):
    namespace = namespace_api.get(context, namespace_name, session)
    values.update({'namespace_id': namespace['id']})

    md_object = models.MetadefObject()
    metadef_utils.drop_protected_attrs(models.MetadefObject, values)
    md_object.update(values.copy())
    try:
        md_object.save(session=session)
    except db_exc.DBDuplicateEntry:
        LOG.debug("A metadata definition object with name=%(name)s"
                  " in namespace=%(namespace_name)s already exists.",
                  {'name': md_object.name,
                   'namespace_name': namespace_name})
        raise exc.MetadefDuplicateObject(
            object_name=md_object.name, namespace_name=namespace_name)

    return md_object.to_dict()


def get(context, namespace_name, name, session):
    md_object = _get_by_name(context, namespace_name, name, session)

    return md_object.to_dict()


def update(context, namespace_name, object_id, values, session):
    """Update an object, raise if ns not found/visible or duplicate result"""
    namespace_api.get(context, namespace_name, session)

    md_object = _get(context, object_id, session)
    metadef_utils.drop_protected_attrs(models.MetadefObject, values)
    # values['updated_at'] = timeutils.utcnow() - done by TS mixin
    try:
        md_object.update(values.copy())
        md_object.save(session=session)
    except db_exc.DBDuplicateEntry:
        LOG.debug("Invalid update. It would result in a duplicate"
                  " metadata definition object with same name=%(name)s"
                  " in namespace=%(namespace_name)s.",
                  {'name': md_object.name, 'namespace_name': namespace_name})
        emsg = (_("Invalid update. It would result in a duplicate"
                  " metadata definition object with the same name=%(name)s"
                  " in namespace=%(namespace_name)s.")
                % {'name': md_object.name, 'namespace_name': namespace_name})
        raise exc.MetadefDuplicateObject(emsg)

    return md_object.to_dict()


def delete(context, namespace_name, object_name, session):
    namespace_api.get(context, namespace_name, session)
    md_object = _get_by_name(context, namespace_name, object_name, session)

    session.delete(md_object)
    session.flush()

    return md_object.to_dict()


def delete_namespace_content(context, namespace_id, session):
    """Use this def only if the ns for the id has been verified as visible"""

    count = 0
    query = session.query(models.MetadefObject).filter_by(
        namespace_id=namespace_id)
    count = query.delete(synchronize_session='fetch')
    return count


def delete_by_namespace_name(context, namespace_name, session):
    namespace = namespace_api.get(context, namespace_name, session)
    return delete_namespace_content(context, namespace['id'], session)


def count(context, namespace_name, session):
    """Get the count of objects for a namespace, raise if ns not found"""
    namespace = namespace_api.get(context, namespace_name, session)

    query = session.query(func.count(models.MetadefObject.id)).filter_by(
        namespace_id=namespace['id'])
    return query.scalar()



# Copyright 2022 OVHCloud
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import urllib.request

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import encodeutils
from oslo_utils import excutils
from taskflow.patterns import linear_flow as lf

from glance.async_.flows._internal_plugins import base_download
from glance.async_ import utils
from glance.common import exception
from glance.common import utils as common_utils
from glance.i18n import _, _LI, _LE

LOG = logging.getLogger(__name__)

CONF = cfg.CONF


class _DownloadGlanceImage(base_download.BaseDownload):

    def __init__(self, context, task_id, task_type, action_wrapper, stores,
                 glance_region, glance_image_id, glance_service_interface):
        self.context = context
        self.glance_region = glance_region
        self.glance_image_id = glance_image_id
        self.glance_service_interface = glance_service_interface
        super(_DownloadGlanceImage,
              self).__init__(task_id, task_type, action_wrapper, stores,
                             'GlanceDownload')

    def execute(self, image_size):
        """Create temp file into store and return path to it

        :param image_size: Glance Image Size retrieved from ImportMetadata task
        """
        try:
            glance_endpoint = utils.get_glance_endpoint(
                self.context,
                self.glance_region,
                self.glance_service_interface)
            image_download_url = '%s/v2/images/%s/file' % (
                glance_endpoint, self.glance_image_id)
            if not common_utils.validate_import_uri(image_download_url):
                LOG.debug("Processed URI for glance-download does not pass "
                          "filtering: %s", image_download_url)
                msg = (_("Processed URI for glance-download does not pass "
                         "filtering: %s") % image_download_url)
                raise exception.ImportTaskError(msg)
            LOG.info(_LI("Downloading glance image %s"), image_download_url)
            token = self.context.auth_token
            request = urllib.request.Request(image_download_url,
                                             headers={'X-Auth-Token': token})
            data = urllib.request.urlopen(request)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error(
                    _LE("Task %(task_id)s failed with exception %(error)s"), {
                        "error": encodeutils.exception_to_unicode(e),
                        "task_id": self.task_id
                    })

        self._path, bytes_written = self.store.add(self.image_id, data, 0)[0:2]
        if bytes_written != image_size:
            msg = (_("Task %(task_id)s failed because downloaded data "
                     "size %(data_size)i is different from expected %("
                     "expected)i") %
                   {"task_id": self.task_id, "data_size": bytes_written,
                    "expected": image_size})
            raise exception.ImportTaskError(msg)
        return self._path


def get_flow(**kwargs):
    """Return task flow for no-op.

    :param context: request context
    :param task_id: Task ID.
    :param task_type: Type of the task.
    :param image_repo: Image repository used.
    :param image_id: Image ID
    :param source_region: Source region name
    """
    context = kwargs.get('context')
    task_id = kwargs.get('task_id')
    task_type = kwargs.get('task_type')
    action_wrapper = kwargs.get('action_wrapper')
    stores = kwargs.get('backend', [None])
    # glance-download parameters
    import_req = kwargs.get('import_req')
    method = import_req.get('method')
    glance_region = method.get('glance_region')
    glance_image_id = method.get('glance_image_id')
    glance_service_interface = method.get('glance_service_interface')

    return lf.Flow(task_type).add(
        _DownloadGlanceImage(context, task_id, task_type, action_wrapper,
                             stores, glance_region, glance_image_id,
                             glance_service_interface),
    )



# Copyright 2015 Hewlett Packard
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests healthcheck middleware."""

import http.client
import tempfile

import httplib2

from glance.tests import functional
from glance.tests import utils


class HealthcheckMiddlewareTest(functional.FunctionalTest):

    def request(self):
        url = 'http://127.0.0.1:%s/healthcheck' % self.api_port
        http = httplib2.Http()
        return http.request(url, 'GET')

    @utils.skip_if_disabled
    def test_healthcheck_enabled(self):
        self.cleanup()
        self.start_servers(**self.__dict__.copy())

        response, content = self.request()
        self.assertEqual(b'OK', content)
        self.assertEqual(http.client.OK, response.status)

        self.stop_servers()

    def test_healthcheck_disabled(self):
        with tempfile.NamedTemporaryFile() as test_disable_file:
            self.cleanup()
            self.api_server.disable_path = test_disable_file.name
            self.start_servers(**self.__dict__.copy())

            response, content = self.request()
            self.assertEqual(b'DISABLED BY FILE', content)
            self.assertEqual(http.client.SERVICE_UNAVAILABLE, response.status)

            self.stop_servers()



# Copyright 2021 Red Hat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

from glance.api.v2 import cached_images
from glance import notifier
import glance.tests.unit.utils as unit_test_utils
import glance.tests.utils as test_utils


UUID1 = 'c80a1a6c-bd1f-41c5-90ee-81afedb1d58d'


class FakeImage(object):
    def __init__(self, id=None, status='active', container_format='ami',
                 disk_format='ami', locations=None):
        self.id = id or UUID1
        self.status = status
        self.container_format = container_format
        self.disk_format = disk_format
        self.locations = locations
        self.owner = unit_test_utils.TENANT1
        self.created_at = ''
        self.updated_at = ''
        self.min_disk = ''
        self.min_ram = ''
        self.protected = False
        self.checksum = ''
        self.os_hash_algo = ''
        self.os_hash_value = ''
        self.size = 0
        self.virtual_size = 0
        self.visibility = 'public'
        self.os_hidden = False
        self.name = 'foo'
        self.tags = []
        self.extra_properties = {}
        self.member = self.owner

        # NOTE(danms): This fixture looks more like the db object than
        # the proxy model. This needs fixing all through the tests
        # below.
        self.image_id = self.id


class TestCacheManageAPI(test_utils.BaseTestCase):

    def setUp(self):
        super(TestCacheManageAPI, self).setUp()
        self.req = unit_test_utils.get_fake_request()

    def _main_test_helper(self, argv, status='active', image_mock=True):
        with mock.patch.object(notifier.ImageRepoProxy,
                               'get') as mock_get:
            image = FakeImage(status=status)
            mock_get.return_value = image
            with mock.patch.object(cached_images.CacheController,
                                   '_enforce') as e:
                with mock.patch('glance.image_cache.ImageCache') as ic:
                    cc = cached_images.CacheController()
                    cc.cache = ic
                    c_calls = []
                    c_calls += argv[0].split(',')
                    for call in c_calls:
                        mock.patch.object(ic, call)
                    test_call = getattr(cc, argv[1])
                    new_policy = argv[2]
                    args = []
                    if len(argv) == 4:
                        args = argv[3:]
                    test_call(self.req, *args)
                    if image_mock:
                        e.assert_called_once_with(self.req, image=image,
                                                  new_policy=new_policy)
                    else:
                        e.assert_called_once_with(self.req,
                                                  new_policy=new_policy)
                    mcs = []
                    for method in ic.method_calls:
                        mcs.append(str(method))
                    for call in c_calls:
                        if args == []:
                            args.append("")
                        elif args[0] and not args[0].endswith("'"):
                            args[0] = "'" + args[0] + "'"
                        self.assertIn("call." + call + "(" + args[0] + ")",
                                      mcs)
                    self.assertEqual(len(c_calls), len(mcs))

    def test_delete_cache_entry(self):
        self._main_test_helper(['delete_cached_image,delete_queued_image',
                                'delete_cache_entry',
                                'cache_delete',
                                UUID1])

    def test_clear_cache(self):
        self._main_test_helper(
            ['delete_all_cached_images,delete_all_queued_images',
             'clear_cache',
             'cache_delete'], image_mock=False)

    def test_get_cache_state(self):
        self._main_test_helper(['get_cached_images,get_queued_images',
                                'get_cache_state',
                                'cache_list'], image_mock=False)

    @mock.patch.object(cached_images, 'WORKER')
    def test_queue_image_from_api(self, mock_worker):
        self._main_test_helper(['queue_image',
                                'queue_image_from_api',
                                'cache_image',
                                UUID1])
        mock_worker.submit.assert_called_once_with(UUID1)

    def test_init_no_config(self):
        # Make sure the worker was reset to uninitialized
        self.assertIsNone(cached_images.WORKER)
        self.config(image_cache_dir=None)
        cached_images.CacheController()

        # Make sure it is still None because image_cache_dir was not
        # set
        self.assertIsNone(cached_images.WORKER)

    def test_init_with_config(self):
        # Make sure the worker was reset to uninitialized
        self.assertIsNone(cached_images.WORKER)
        self.config(image_cache_dir='/tmp')
        cached_images.CacheController()

        # Make sure we initialized it because config told us to
        self.assertIsNotNone(cached_images.WORKER)
        self.assertTrue(cached_images.WORKER.is_alive())
        cached_images.WORKER.terminate()


class TestCacheWorker(test_utils.BaseTestCase):
    @mock.patch('glance.image_cache.prefetcher.Prefetcher')
    def test_worker_lifecycle(self, mock_pf):
        worker = cached_images.CacheWorker()
        self.assertFalse(worker.is_alive())
        worker.start()
        self.assertTrue(worker.is_alive())
        worker.submit('123')
        worker.submit('456')
        self.assertTrue(worker.is_alive())
        worker.terminate()
        self.assertFalse(worker.is_alive())
        mock_pf.return_value.fetch_image_into_cache.assert_has_calls([
            mock.call('123'), mock.call('456')])



# Copyright (c) 2014 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re

from hacking import core

"""
Guidelines for writing new hacking checks

 - Use only for Glance-specific tests. OpenStack general tests
   should be submitted to the common 'hacking' module.
 - Pick numbers in the range G3xx. Find the current test with
   the highest allocated number and then pick the next value.
   If nova has an N3xx code for that test, use the same number.
 - Keep the test method code in the source file ordered based
   on the G3xx value.
 - List the new rule in the top level HACKING.rst file
 - Add test cases for each new rule to glance/tests/test_hacking.py

"""


asse_trueinst_re = re.compile(
    r"(.)*assertTrue\(isinstance\((\w|\.|\'|\"|\[|\])+, "
    r"(\w|\.|\'|\"|\[|\])+\)\)")
asse_equal_type_re = re.compile(
    r"(.)*assertEqual\(type\((\w|\.|\'|\"|\[|\])+\), "
    r"(\w|\.|\'|\"|\[|\])+\)")
asse_equal_end_with_none_re = re.compile(
    r"(.)*assertEqual\((\w|\.|\'|\"|\[|\])+, None\)")
asse_equal_start_with_none_re = re.compile(
    r"(.)*assertEqual\(None, (\w|\.|\'|\"|\[|\])+\)")
unicode_func_re = re.compile(r"(\s|\W|^)unicode\(")
dict_constructor_with_list_copy_re = re.compile(r".*\bdict\((\[)?(\(|\[)")


@core.flake8ext
def assert_true_instance(logical_line):
    """Check for assertTrue(isinstance(a, b)) sentences

    G316
    """
    if asse_trueinst_re.match(logical_line):
        yield (0, "G316: assertTrue(isinstance(a, b)) sentences not allowed")


@core.flake8ext
def assert_equal_type(logical_line):
    """Check for assertEqual(type(A), B) sentences

    G317
    """
    if asse_equal_type_re.match(logical_line):
        yield (0, "G317: assertEqual(type(A), B) sentences not allowed")


@core.flake8ext
def assert_equal_none(logical_line):
    """Check for assertEqual(A, None) or assertEqual(None, A) sentences

    G318
    """
    res = (asse_equal_start_with_none_re.match(logical_line) or
           asse_equal_end_with_none_re.match(logical_line))
    if res:
        yield (0, "G318: assertEqual(A, None) or assertEqual(None, A) "
               "sentences not allowed")


@core.flake8ext
def no_translate_debug_logs(logical_line, filename):
    dirs = [
        "glance/api",
        "glance/cmd",
        "glance/common",
        "glance/db",
        "glance/domain",
        "glance/image_cache",
        "glance/quota",
        "glance/store",
        "glance/tests",
    ]

    if max([name in filename for name in dirs]):
        if logical_line.startswith("LOG.debug(_("):
            yield(0, "G319: Don't translate debug level logs")


@core.flake8ext
def check_no_contextlib_nested(logical_line):
    msg = ("G327: contextlib.nested is deprecated since Python 2.7. See "
           "https://docs.python.org/2/library/contextlib.html#contextlib."
           "nested for more information.")
    if ("with contextlib.nested(" in logical_line or
            "with nested(" in logical_line):
        yield(0, msg)


@core.flake8ext
def dict_constructor_with_list_copy(logical_line):
    msg = ("G328: Must use a dict comprehension instead of a dict constructor "
           "with a sequence of key-value pairs.")
    if dict_constructor_with_list_copy_re.match(logical_line):
        yield (0, msg)


@core.flake8ext
def no_log_warn(logical_line):
    """Disallow 'LOG.warn('

    Use LOG.warning() instead of Deprecated LOG.warn().
    https://docs.python.org/3/library/logging.html#logging.warning
    """

    msg = ("G330: LOG.warn is deprecated, please use LOG.warning!")
    if "LOG.warn(" in logical_line:
        yield (0, msg)



# Copyright 2011-2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import http.client
import os
import sys
import time

import httplib2
from oslo_config import cfg
from oslo_serialization import jsonutils
from oslo_utils.fixture import uuidsentinel as uuids

from glance import context
import glance.db as db_api
from glance.tests import functional
from glance.tests.utils import execute

CONF = cfg.CONF


class TestScrubber(functional.FunctionalTest):

    """Test that delayed_delete works and the scrubber deletes"""

    def setUp(self):
        super(TestScrubber, self).setUp()
        self.api_server.deployment_flavor = 'noauth'
        self.api_server.send_identity_credentials = True
        self.admin_context = context.get_admin_context(show_deleted=True)
        CONF.set_override('sql_connection', self.api_server.sql_connection)

    def _headers(self, custom_headers=None):
        base_headers = {
            'X-Identity-Status': 'Confirmed',
            'X-Auth-Token': '932c5c84-02ac-4fe5-a9ba-620af0e2bb96',
            'X-User-Id': 'f9a41d13-0c13-47e9-bee2-ce4e8bfe958e',
            'X-Tenant-Id': uuids.TENANT1,
            'X-Roles': 'reader,member',
        }
        base_headers.update(custom_headers or {})
        return base_headers

    def _send_create_image_http_request(self, path, body=None):
        headers = {
            "Content-Type": "application/json",
            "X-Roles": "admin",
        }
        body = body or {'container_format': 'ovf',
                        'disk_format': 'raw',
                        'name': 'test_image',
                        'visibility': 'public'}
        body = jsonutils.dumps(body)
        return httplib2.Http().request(path, 'POST', body,
                                       self._headers(headers))

    def _send_upload_image_http_request(self, path, body=None):
        headers = {
            "Content-Type": "application/octet-stream"
        }
        return httplib2.Http().request(path, 'PUT', body,
                                       self._headers(headers))

    def _send_http_request(self, path, method):
        headers = {
            "Content-Type": "application/json"
        }
        return httplib2.Http().request(path, method, None,
                                       self._headers(headers))

    def _get_pending_delete_image(self, image_id):
        # In Glance V2, there is no way to get the 'pending_delete' image from
        # API. So we get the image from db here for testing.
        # Clean the session cache first to avoid connecting to the old db data.
        db_api.get_api()._FACADE = None
        image = db_api.get_api().image_get(self.admin_context, image_id)
        return image

    def test_delayed_delete(self):
        """
        test that images don't get deleted immediately and that the scrubber
        scrubs them
        """
        self.cleanup()
        kwargs = self.__dict__.copy()
        self.start_servers(delayed_delete=True, daemon=True,
                           metadata_encryption_key='', **kwargs)
        path = "http://%s:%d/v2/images" % ("127.0.0.1", self.api_port)
        response, content = self._send_create_image_http_request(path)
        self.assertEqual(http.client.CREATED, response.status)
        image = jsonutils.loads(content)
        self.assertEqual('queued', image['status'])

        file_path = "%s/%s/file" % (path, image['id'])
        response, content = self._send_upload_image_http_request(file_path,
                                                                 body='XXX')
        self.assertEqual(http.client.NO_CONTENT, response.status)

        path = "%s/%s" % (path, image['id'])
        response, content = self._send_http_request(path, 'GET')
        image = jsonutils.loads(content)
        self.assertEqual('active', image['status'])

        response, content = self._send_http_request(path, 'DELETE')
        self.assertEqual(http.client.NO_CONTENT, response.status)

        image = self._get_pending_delete_image(image['id'])
        self.assertEqual('pending_delete', image['status'])

        self.wait_for_scrub(image['id'])

        self.stop_servers()

    def test_scrubber_app(self):
        """
        test that the glance-scrubber script runs successfully when not in
        daemon mode
        """
        self.cleanup()
        kwargs = self.__dict__.copy()
        self.start_servers(delayed_delete=True, daemon=False,
                           metadata_encryption_key='', **kwargs)
        path = "http://%s:%d/v2/images" % ("127.0.0.1", self.api_port)
        response, content = self._send_create_image_http_request(path)
        self.assertEqual(http.client.CREATED, response.status)
        image = jsonutils.loads(content)
        self.assertEqual('queued', image['status'])

        file_path = "%s/%s/file" % (path, image['id'])
        response, content = self._send_upload_image_http_request(file_path,
                                                                 body='XXX')
        self.assertEqual(http.client.NO_CONTENT, response.status)

        path = "%s/%s" % (path, image['id'])
        response, content = self._send_http_request(path, 'GET')
        image = jsonutils.loads(content)
        self.assertEqual('active', image['status'])

        response, content = self._send_http_request(path, 'DELETE')
        self.assertEqual(http.client.NO_CONTENT, response.status)

        image = self._get_pending_delete_image(image['id'])
        self.assertEqual('pending_delete', image['status'])

        # wait for the scrub time on the image to pass
        time.sleep(self.api_server.scrub_time)

        # scrub images and make sure they get deleted
        exe_cmd = "%s -m glance.cmd.scrubber" % sys.executable
        cmd = ("%s --config-file %s" %
               (exe_cmd, self.scrubber_daemon.conf_file_name))
        exitcode, out, err = execute(cmd, raise_error=False)
        self.assertEqual(0, exitcode)

        self.wait_for_scrub(image['id'])

        self.stop_servers()

    def test_scrubber_delete_handles_exception(self):
        """
        Test that the scrubber handles the case where an
        exception occurs when _delete() is called. The scrubber
        should not write out queue files in this case.
        """

        # Start servers.
        self.cleanup()
        kwargs = self.__dict__.copy()
        self.start_servers(delayed_delete=True, daemon=False,
                           default_store='file', **kwargs)

        # Check that we are using a file backend.
        self.assertEqual(self.api_server.default_store, 'file')

        # add an image
        path = "http://%s:%d/v2/images" % ("127.0.0.1", self.api_port)
        response, content = self._send_create_image_http_request(path)
        self.assertEqual(http.client.CREATED, response.status)
        image = jsonutils.loads(content)
        self.assertEqual('queued', image['status'])

        file_path = "%s/%s/file" % (path, image['id'])
        response, content = self._send_upload_image_http_request(file_path,
                                                                 body='XXX')
        self.assertEqual(http.client.NO_CONTENT, response.status)

        path = "%s/%s" % (path, image['id'])
        response, content = self._send_http_request(path, 'GET')
        image = jsonutils.loads(content)
        self.assertEqual('active', image['status'])
        # delete the image
        response, content = self._send_http_request(path, 'DELETE')
        self.assertEqual(http.client.NO_CONTENT, response.status)
        # ensure the image is marked pending delete.
        image = self._get_pending_delete_image(image['id'])
        self.assertEqual('pending_delete', image['status'])

        # Remove the file from the backend.
        file_path = os.path.join(self.api_server.image_dir, image['id'])
        os.remove(file_path)

        # Wait for the scrub time on the image to pass
        time.sleep(self.api_server.scrub_time)

        # run the scrubber app, and ensure it doesn't fall over
        exe_cmd = "%s -m glance.cmd.scrubber" % sys.executable
        cmd = ("%s --config-file %s" %
               (exe_cmd, self.scrubber_daemon.conf_file_name))
        exitcode, out, err = execute(cmd, raise_error=False)
        self.assertEqual(0, exitcode)

        self.wait_for_scrub(image['id'])

        self.stop_servers()

    def test_scrubber_app_queue_errors_not_daemon(self):
        """
        test that the glance-scrubber exits with an exit code > 0 when it
        fails to lookup images, indicating a configuration error when not
        in daemon mode.

        Related-Bug: #1548289
        """
        # Don't start the registry server to cause intended failure
        # Don't start the api server to save time
        exitcode, out, err = self.scrubber_daemon.start(
            delayed_delete=True, daemon=False)
        self.assertEqual(0, exitcode,
                         "Failed to spin up the Scrubber daemon. "
                         "Got: %s" % err)

        # Run the Scrubber
        exe_cmd = "%s -m glance.cmd.scrubber" % sys.executable
        cmd = ("%s --config-file %s" %
               (exe_cmd, self.scrubber_daemon.conf_file_name))
        exitcode, out, err = execute(cmd, raise_error=False)

        self.assertEqual(1, exitcode)
        self.assertIn('Can not get scrub jobs from queue', str(err))

        self.stop_server(self.scrubber_daemon)

    def test_scrubber_restore_image(self):
        self.cleanup()
        kwargs = self.__dict__.copy()
        self.start_servers(delayed_delete=True, daemon=False,
                           metadata_encryption_key='', **kwargs)
        path = "http://%s:%d/v2/images" % ("127.0.0.1", self.api_port)
        response, content = self._send_create_image_http_request(path)
        self.assertEqual(http.client.CREATED, response.status)
        image = jsonutils.loads(content)
        self.assertEqual('queued', image['status'])

        file_path = "%s/%s/file" % (path, image['id'])
        response, content = self._send_upload_image_http_request(file_path,
                                                                 body='XXX')
        self.assertEqual(http.client.NO_CONTENT, response.status)

        path = "%s/%s" % (path, image['id'])
        response, content = self._send_http_request(path, 'GET')
        image = jsonutils.loads(content)
        self.assertEqual('active', image['status'])

        response, content = self._send_http_request(path, 'DELETE')
        self.assertEqual(http.client.NO_CONTENT, response.status)

        image = self._get_pending_delete_image(image['id'])
        self.assertEqual('pending_delete', image['status'])

        def _test_content():
            exe_cmd = "%s -m glance.cmd.scrubber" % sys.executable
            cmd = ("%s --config-file %s --restore %s" %
                   (exe_cmd, self.scrubber_daemon.conf_file_name, image['id']))
            return execute(cmd, raise_error=False)

        exitcode, out, err = self.wait_for_scrubber_shutdown(_test_content)
        self.assertEqual(0, exitcode)

        response, content = self._send_http_request(path, 'GET')
        image = jsonutils.loads(content)
        self.assertEqual('active', image['status'])

        self.stop_servers()

    def test_scrubber_restore_active_image_raise_error(self):
        self.cleanup()
        self.start_servers(delayed_delete=True, daemon=False,
                           metadata_encryption_key='')

        path = "http://%s:%d/v2/images" % ("127.0.0.1", self.api_port)
        response, content = self._send_create_image_http_request(path)
        self.assertEqual(http.client.CREATED, response.status)
        image = jsonutils.loads(content)
        self.assertEqual('queued', image['status'])

        file_path = "%s/%s/file" % (path, image['id'])
        response, content = self._send_upload_image_http_request(file_path,
                                                                 body='XXX')
        self.assertEqual(http.client.NO_CONTENT, response.status)

        path = "%s/%s" % (path, image['id'])
        response, content = self._send_http_request(path, 'GET')
        image = jsonutils.loads(content)
        self.assertEqual('active', image['status'])

        def _test_content():
            exe_cmd = "%s -m glance.cmd.scrubber" % sys.executable
            cmd = ("%s --config-file %s --restore %s" %
                   (exe_cmd, self.scrubber_daemon.conf_file_name, image['id']))
            return execute(cmd, raise_error=False)

        exitcode, out, err = self.wait_for_scrubber_shutdown(_test_content)
        self.assertEqual(1, exitcode)
        self.assertIn('cannot restore the image from active to active '
                      '(wanted from_state=pending_delete)', str(err))

        self.stop_servers()

    def test_scrubber_restore_image_non_exist(self):

        def _test_content():
            scrubber = functional.ScrubberDaemon(self.test_dir,
                                                 self.policy_file)
            scrubber.write_conf(daemon=False)
            scrubber.needs_database = True
            scrubber.create_database()
            exe_cmd = "%s -m glance.cmd.scrubber" % sys.executable
            cmd = ("%s --config-file %s --restore fake_image_id" %
                   (exe_cmd, scrubber.conf_file_name))
            return execute(cmd, raise_error=False)

        exitcode, out, err = self.wait_for_scrubber_shutdown(_test_content)
        self.assertEqual(1, exitcode)
        self.assertIn('No image found with ID fake_image_id', str(err))

    def test_scrubber_restore_image_with_daemon_raise_error(self):

        exe_cmd = "%s -m glance.cmd.scrubber" % sys.executable
        cmd = ("%s --daemon --restore fake_image_id" % exe_cmd)
        exitcode, out, err = execute(cmd, raise_error=False)

        self.assertEqual(1, exitcode)
        self.assertIn('The restore and daemon options should not be set '
                      'together', str(err))

    def test_scrubber_restore_image_with_daemon_running(self):
        self.cleanup()
        self.scrubber_daemon.start(daemon=True)
        # Give the scrubber some time to start.
        time.sleep(5)

        exe_cmd = "%s -m glance.cmd.scrubber" % sys.executable
        cmd = ("%s --restore fake_image_id" % exe_cmd)
        exitcode, out, err = execute(cmd, raise_error=False)
        self.assertEqual(1, exitcode)
        self.assertIn('glance-scrubber is already running', str(err))

        self.stop_server(self.scrubber_daemon)

    def wait_for_scrubber_shutdown(self, func):
        # NOTE(wangxiyuan, rosmaita): The image-restore functionality contains
        # a check to make sure the scrubber isn't also running in daemon mode
        # to prevent a race condition between a delete and a restore.
        # Sometimes the glance-scrubber process which is setup by the
        # previous test can't be shutdown immediately, so if we get the "daemon
        # running" message we sleep and try again.
        not_down_msg = 'glance-scrubber is already running'
        total_wait = 15
        for _ in range(total_wait):
            exitcode, out, err = func()
            if exitcode == 1 and not_down_msg in str(err):
                time.sleep(1)
                continue
            return exitcode, out, err
        else:
            self.fail('Scrubber did not shut down within {} sec'.format(
                total_wait))

    def wait_for_scrub(self, image_id):
        """
        NOTE(jkoelker) The build servers sometimes take longer than 15 seconds
        to scrub. Give it up to 5 min, checking checking every 15 seconds.
        When/if it flips to deleted, bail immediately.
        """
        wait_for = 300    # seconds
        check_every = 15  # seconds
        for _ in range(wait_for // check_every):
            time.sleep(check_every)
            image = db_api.get_api().image_get(self.admin_context, image_id)
            if (image['status'] == 'deleted' and
                    image['deleted'] == True):
                break
            else:
                continue
        else:
            self.fail('image was never scrubbed')



# Copyright 2013, Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import fixtures
from unittest import mock
from unittest.mock import patch
import uuid

from oslo_limit import exception as ol_exc
from oslo_utils import encodeutils
from oslo_utils import units

from glance.common import exception
from glance.common import store_utils
import glance.quota
from glance.quota import keystone as ks_quota
from glance.tests.unit import fixtures as glance_fixtures
from glance.tests.unit import utils as unit_test_utils
from glance.tests import utils as test_utils

UUID1 = 'c80a1a6c-bd1f-41c5-90ee-81afedb1d58d'


class FakeContext(object):
    owner = 'someone'
    is_admin = False


class FakeImage(object):
    size = None
    image_id = 'someid'
    locations = [{'url': 'file:///not/a/path', 'metadata': {}}]
    tags = set([])

    def set_data(self, data, size=None, backend=None, set_active=True):
        self.size = 0
        for d in data:
            self.size += len(d)

    def __init__(self, **kwargs):
        self.extra_properties = kwargs.get('extra_properties', {})


class TestImageQuota(test_utils.BaseTestCase):

    def setUp(self):
        super(TestImageQuota, self).setUp()

    def _get_image(self, location_count=1, image_size=10):
        context = FakeContext()
        db_api = unit_test_utils.FakeDB()
        store_api = unit_test_utils.FakeStoreAPI()
        store = unit_test_utils.FakeStoreUtils(store_api)
        base_image = FakeImage()
        base_image.image_id = 'xyz'
        base_image.size = image_size
        image = glance.quota.ImageProxy(base_image, context, db_api, store)
        locations = []
        for i in range(location_count):
            locations.append({'url': 'file:///g/there/it/is%d' % i,
                              'metadata': {}, 'status': 'active'})
        image_values = {'id': 'xyz', 'owner': context.owner,
                        'status': 'active', 'size': image_size,
                        'locations': locations}
        db_api.image_create(context, image_values)
        return image

    def test_quota_allowed(self):
        quota = 10
        self.config(user_storage_quota=str(quota))
        context = FakeContext()
        db_api = unit_test_utils.FakeDB()
        store_api = unit_test_utils.FakeStoreAPI()
        store = unit_test_utils.FakeStoreUtils(store_api)
        base_image = FakeImage()
        base_image.image_id = 'id'
        image = glance.quota.ImageProxy(base_image, context, db_api, store)
        data = '*' * quota
        base_image.set_data(data, size=None)
        image.set_data(data)
        self.assertEqual(quota, base_image.size)

    def _test_quota_allowed_unit(self, data_length, config_quota):
        self.config(user_storage_quota=config_quota)
        context = FakeContext()
        db_api = unit_test_utils.FakeDB()
        store_api = unit_test_utils.FakeStoreAPI()
        store = unit_test_utils.FakeStoreUtils(store_api)
        base_image = FakeImage()
        base_image.image_id = 'id'
        image = glance.quota.ImageProxy(base_image, context, db_api, store)
        data = '*' * data_length
        base_image.set_data(data, size=None)
        image.set_data(data)
        self.assertEqual(data_length, base_image.size)

    def test_quota_allowed_unit_b(self):
        self._test_quota_allowed_unit(10, '10B')

    def test_quota_allowed_unit_kb(self):
        self._test_quota_allowed_unit(10, '1KB')

    def test_quota_allowed_unit_mb(self):
        self._test_quota_allowed_unit(10, '1MB')

    def test_quota_allowed_unit_gb(self):
        self._test_quota_allowed_unit(10, '1GB')

    def test_quota_allowed_unit_tb(self):
        self._test_quota_allowed_unit(10, '1TB')

    def _quota_exceeded_size(self, quota, data,
                             deleted=True, size=None):
        self.config(user_storage_quota=quota)
        context = FakeContext()
        db_api = unit_test_utils.FakeDB()
        store_api = unit_test_utils.FakeStoreAPI()
        store = unit_test_utils.FakeStoreUtils(store_api)
        base_image = FakeImage()
        base_image.image_id = 'id'
        image = glance.quota.ImageProxy(base_image, context, db_api, store)

        if deleted:
            with patch.object(store_utils, 'safe_delete_from_backend'):
                store_utils.safe_delete_from_backend(
                    context,
                    image.image_id,
                    base_image.locations[0])

        self.assertRaises(exception.StorageQuotaFull,
                          image.set_data,
                          data,
                          size=size)

    def test_quota_exceeded_no_size(self):
        quota = 10
        data = '*' * (quota + 1)
        # NOTE(jbresnah) When the image size is None it means that it is
        # not known.  In this case the only time we will raise an
        # exception is when there is no room left at all, thus we know
        # it will not fit.
        # That's why 'get_remaining_quota' is mocked with return_value = 0.
        with patch.object(glance.api.common, 'get_remaining_quota',
                          return_value=0):
            self._quota_exceeded_size(str(quota), data)

    def test_quota_exceeded_with_right_size(self):
        quota = 10
        data = '*' * (quota + 1)
        self._quota_exceeded_size(str(quota), data, size=len(data),
                                  deleted=False)

    def test_quota_exceeded_with_right_size_b(self):
        quota = 10
        data = '*' * (quota + 1)
        self._quota_exceeded_size('10B', data, size=len(data),
                                  deleted=False)

    def test_quota_exceeded_with_right_size_kb(self):
        quota = units.Ki
        data = '*' * (quota + 1)
        self._quota_exceeded_size('1KB', data, size=len(data),
                                  deleted=False)

    def test_quota_exceeded_with_lie_size(self):
        quota = 10
        data = '*' * (quota + 1)
        self._quota_exceeded_size(str(quota), data, deleted=False,
                                  size=quota - 1)

    def test_quota_exceeded_keystone_quotas(self):
        # Set our global limit to a tiny ten bytes
        self.config(user_storage_quota='10B')
        context = FakeContext()
        db_api = unit_test_utils.FakeDB()
        store_api = unit_test_utils.FakeStoreAPI()
        store = unit_test_utils.FakeStoreUtils(store_api)
        base_image = FakeImage()
        base_image.image_id = 'id'
        image = glance.quota.ImageProxy(base_image, context, db_api, store)

        # With keystone quotas disabled, a 100 byte image should fail the
        # global limit.
        data = '*' * 100
        self.assertRaises(exception.StorageQuotaFull,
                          image.set_data,
                          data,
                          size=len(data))

        # If we turn on keystone quotas, the global limit gets ignored
        # so the same image no longer fails.
        self.config(use_keystone_limits=True)
        image.set_data(data, size=len(data))

    def test_append_location(self):
        new_location = {'url': 'file:///a/path', 'metadata': {},
                        'status': 'active'}
        image = self._get_image()
        pre_add_locations = image.locations[:]
        image.locations.append(new_location)
        pre_add_locations.append(new_location)
        self.assertEqual(image.locations, pre_add_locations)

    def test_insert_location(self):
        new_location = {'url': 'file:///a/path', 'metadata': {},
                        'status': 'active'}
        image = self._get_image()
        pre_add_locations = image.locations[:]
        image.locations.insert(0, new_location)
        pre_add_locations.insert(0, new_location)
        self.assertEqual(image.locations, pre_add_locations)

    def test_extend_location(self):
        new_location = {'url': 'file:///a/path', 'metadata': {},
                        'status': 'active'}
        image = self._get_image()
        pre_add_locations = image.locations[:]
        image.locations.extend([new_location])
        pre_add_locations.extend([new_location])
        self.assertEqual(image.locations, pre_add_locations)

    def test_iadd_location(self):
        new_location = {'url': 'file:///a/path', 'metadata': {},
                        'status': 'active'}
        image = self._get_image()
        pre_add_locations = image.locations[:]
        image.locations += [new_location]
        pre_add_locations += [new_location]
        self.assertEqual(image.locations, pre_add_locations)

    def test_set_location(self):
        new_location = {'url': 'file:///a/path', 'metadata': {},
                        'status': 'active'}
        image = self._get_image()
        image.locations = [new_location]
        self.assertEqual(image.locations, [new_location])

    def _make_image_with_quota(self, image_size=10, location_count=2):
        quota = image_size * location_count
        self.config(user_storage_quota=str(quota))
        return self._get_image(image_size=image_size,
                               location_count=location_count)

    def test_exceed_append_location(self):
        image = self._make_image_with_quota()
        self.assertRaises(exception.StorageQuotaFull,
                          image.locations.append,
                          {'url': 'file:///a/path', 'metadata': {},
                           'status': 'active'})

    def test_exceed_insert_location(self):
        image = self._make_image_with_quota()
        self.assertRaises(exception.StorageQuotaFull,
                          image.locations.insert,
                          0,
                          {'url': 'file:///a/path', 'metadata': {},
                           'status': 'active'})

    def test_exceed_extend_location(self):
        image = self._make_image_with_quota()
        self.assertRaises(exception.StorageQuotaFull,
                          image.locations.extend,
                          [{'url': 'file:///a/path', 'metadata': {},
                            'status': 'active'}])

    def test_set_location_under(self):
        image = self._make_image_with_quota(location_count=1)
        image.locations = [{'url': 'file:///a/path', 'metadata': {},
                            'status': 'active'}]

    def test_set_location_exceed(self):
        image = self._make_image_with_quota(location_count=1)
        try:
            image.locations = [{'url': 'file:///a/path', 'metadata': {},
                                'status': 'active'},
                               {'url': 'file:///a/path2', 'metadata': {},
                                'status': 'active'}]
            self.fail('Should have raised the quota exception')
        except exception.StorageQuotaFull:
            pass

    def test_iadd_location_exceed(self):
        image = self._make_image_with_quota(location_count=1)
        try:
            image.locations += [{'url': 'file:///a/path', 'metadata': {},
                                 'status': 'active'}]
            self.fail('Should have raised the quota exception')
        except exception.StorageQuotaFull:
            pass

    def test_append_location_for_queued_image(self):
        context = FakeContext()
        db_api = unit_test_utils.FakeDB()
        store_api = unit_test_utils.FakeStoreAPI()
        store = unit_test_utils.FakeStoreUtils(store_api)
        base_image = FakeImage()
        base_image.image_id = str(uuid.uuid4())
        image = glance.quota.ImageProxy(base_image, context, db_api, store)
        self.assertIsNone(image.size)

        self.mock_object(store_api, 'get_size_from_backend',
                         unit_test_utils.fake_get_size_from_backend)
        image.locations.append({'url': 'file:///fake.img.tar.gz',
                                'metadata': {}})
        self.assertIn({'url': 'file:///fake.img.tar.gz', 'metadata': {}},
                      image.locations)

    def test_insert_location_for_queued_image(self):
        context = FakeContext()
        db_api = unit_test_utils.FakeDB()
        store_api = unit_test_utils.FakeStoreAPI()
        store = unit_test_utils.FakeStoreUtils(store_api)
        base_image = FakeImage()
        base_image.image_id = str(uuid.uuid4())
        image = glance.quota.ImageProxy(base_image, context, db_api, store)
        self.assertIsNone(image.size)

        self.mock_object(store_api, 'get_size_from_backend',
                         unit_test_utils.fake_get_size_from_backend)
        image.locations.insert(0,
                               {'url': 'file:///fake.img.tar.gz',
                                'metadata': {}})
        self.assertIn({'url': 'file:///fake.img.tar.gz', 'metadata': {}},
                      image.locations)

    def test_set_location_for_queued_image(self):
        context = FakeContext()
        db_api = unit_test_utils.FakeDB()
        store_api = unit_test_utils.FakeStoreAPI()
        store = unit_test_utils.FakeStoreUtils(store_api)
        base_image = FakeImage()
        base_image.image_id = str(uuid.uuid4())
        image = glance.quota.ImageProxy(base_image, context, db_api, store)
        self.assertIsNone(image.size)

        self.mock_object(store_api, 'get_size_from_backend',
                         unit_test_utils.fake_get_size_from_backend)
        image.locations = [{'url': 'file:///fake.img.tar.gz', 'metadata': {}}]
        self.assertEqual([{'url': 'file:///fake.img.tar.gz', 'metadata': {}}],
                         image.locations)

    def test_iadd_location_for_queued_image(self):
        context = FakeContext()
        db_api = unit_test_utils.FakeDB()
        store_api = unit_test_utils.FakeStoreAPI()
        store = unit_test_utils.FakeStoreUtils(store_api)
        base_image = FakeImage()
        base_image.image_id = str(uuid.uuid4())
        image = glance.quota.ImageProxy(base_image, context, db_api, store)
        self.assertIsNone(image.size)

        self.mock_object(store_api, 'get_size_from_backend',
                         unit_test_utils.fake_get_size_from_backend)
        image.locations += [{'url': 'file:///fake.img.tar.gz', 'metadata': {}}]
        self.assertIn({'url': 'file:///fake.img.tar.gz', 'metadata': {}},
                      image.locations)


class TestImagePropertyQuotas(test_utils.BaseTestCase):

    def setUp(self):
        super(TestImagePropertyQuotas, self).setUp()
        self.base_image = FakeImage()
        self.image = glance.quota.ImageProxy(self.base_image,
                                             mock.Mock(),
                                             mock.Mock(),
                                             mock.Mock())

        self.image_repo_mock = mock.Mock()
        self.image_repo_mock.add.return_value = self.base_image
        self.image_repo_mock.save.return_value = self.base_image

        self.image_repo_proxy = glance.quota.ImageRepoProxy(
            self.image_repo_mock,
            mock.Mock(),
            mock.Mock(),
            mock.Mock())

    def test_save_image_with_image_property(self):
        self.config(image_property_quota=1)

        self.image.extra_properties = {'foo': 'bar'}
        self.image_repo_proxy.save(self.image)

        self.image_repo_mock.save.assert_called_once_with(self.base_image,
                                                          from_state=None)

    def test_save_image_too_many_image_properties(self):
        self.config(image_property_quota=1)

        self.image.extra_properties = {'foo': 'bar', 'foo2': 'bar2'}
        exc = self.assertRaises(exception.ImagePropertyLimitExceeded,
                                self.image_repo_proxy.save, self.image)
        self.assertIn("Attempted: 2, Maximum: 1",
                      encodeutils.exception_to_unicode(exc))

    def test_save_image_unlimited_image_properties(self):
        self.config(image_property_quota=-1)

        self.image.extra_properties = {'foo': 'bar'}
        self.image_repo_proxy.save(self.image)

        self.image_repo_mock.save.assert_called_once_with(self.base_image,
                                                          from_state=None)

    def test_add_image_with_image_property(self):
        self.config(image_property_quota=1)

        self.image.extra_properties = {'foo': 'bar'}
        self.image_repo_proxy.add(self.image)

        self.image_repo_mock.add.assert_called_once_with(self.base_image)

    def test_add_image_too_many_image_properties(self):
        self.config(image_property_quota=1)

        self.image.extra_properties = {'foo': 'bar', 'foo2': 'bar2'}
        exc = self.assertRaises(exception.ImagePropertyLimitExceeded,
                                self.image_repo_proxy.add, self.image)
        self.assertIn("Attempted: 2, Maximum: 1",
                      encodeutils.exception_to_unicode(exc))

    def test_add_image_unlimited_image_properties(self):
        self.config(image_property_quota=-1)

        self.image.extra_properties = {'foo': 'bar'}
        self.image_repo_proxy.add(self.image)

        self.image_repo_mock.add.assert_called_once_with(self.base_image)

    def _quota_exceed_setup(self):
        self.config(image_property_quota=2)
        self.base_image.extra_properties = {'foo': 'bar', 'spam': 'ham'}
        self.image = glance.quota.ImageProxy(self.base_image,
                                             mock.Mock(),
                                             mock.Mock(),
                                             mock.Mock())

    def test_modify_image_properties_when_quota_exceeded(self):
        self._quota_exceed_setup()
        self.config(image_property_quota=1)
        self.image.extra_properties = {'foo': 'frob', 'spam': 'eggs'}
        self.image_repo_proxy.save(self.image)
        self.image_repo_mock.save.assert_called_once_with(self.base_image,
                                                          from_state=None)
        self.assertEqual('frob', self.base_image.extra_properties['foo'])
        self.assertEqual('eggs', self.base_image.extra_properties['spam'])

    def test_delete_image_properties_when_quota_exceeded(self):
        self._quota_exceed_setup()
        self.config(image_property_quota=1)
        del self.image.extra_properties['foo']
        self.image_repo_proxy.save(self.image)
        self.image_repo_mock.save.assert_called_once_with(self.base_image,
                                                          from_state=None)
        self.assertNotIn('foo', self.base_image.extra_properties)
        self.assertEqual('ham', self.base_image.extra_properties['spam'])

    def test_invalid_quota_config_parameter(self):
        self.config(user_storage_quota='foo')
        location = {"url": "file:///fake.img.tar.gz", "metadata": {}}
        self.assertRaises(exception.InvalidOptionValue,
                          self.image.locations.append, location)

    def test_exceed_quota_during_patch_operation(self):
        self._quota_exceed_setup()
        self.image.extra_properties['frob'] = 'baz'
        self.image.extra_properties['lorem'] = 'ipsum'
        self.assertEqual('bar', self.base_image.extra_properties['foo'])
        self.assertEqual('ham', self.base_image.extra_properties['spam'])
        self.assertEqual('baz', self.base_image.extra_properties['frob'])
        self.assertEqual('ipsum', self.base_image.extra_properties['lorem'])

        del self.image.extra_properties['frob']
        del self.image.extra_properties['lorem']
        self.image_repo_proxy.save(self.image)
        call_args = mock.call(self.base_image, from_state=None)
        self.assertEqual(call_args, self.image_repo_mock.save.call_args)
        self.assertEqual('bar', self.base_image.extra_properties['foo'])
        self.assertEqual('ham', self.base_image.extra_properties['spam'])
        self.assertNotIn('frob', self.base_image.extra_properties)
        self.assertNotIn('lorem', self.base_image.extra_properties)

    def test_quota_exceeded_after_delete_image_properties(self):
        self.config(image_property_quota=3)
        self.base_image.extra_properties = {'foo': 'bar',
                                            'spam': 'ham',
                                            'frob': 'baz'}
        self.image = glance.quota.ImageProxy(self.base_image,
                                             mock.Mock(),
                                             mock.Mock(),
                                             mock.Mock())
        self.config(image_property_quota=1)
        del self.image.extra_properties['foo']
        self.image_repo_proxy.save(self.image)
        self.image_repo_mock.save.assert_called_once_with(self.base_image,
                                                          from_state=None)
        self.assertNotIn('foo', self.base_image.extra_properties)
        self.assertEqual('ham', self.base_image.extra_properties['spam'])
        self.assertEqual('baz', self.base_image.extra_properties['frob'])


class TestImageTagQuotas(test_utils.BaseTestCase):

    def setUp(self):
        super(TestImageTagQuotas, self).setUp()
        self.base_image = mock.Mock()
        self.base_image.tags = set([])
        self.base_image.extra_properties = {}
        self.image = glance.quota.ImageProxy(self.base_image,
                                             mock.Mock(),
                                             mock.Mock(),
                                             mock.Mock())

        self.image_repo_mock = mock.Mock()
        self.image_repo_proxy = glance.quota.ImageRepoProxy(
            self.image_repo_mock,
            mock.Mock(),
            mock.Mock(),
            mock.Mock())

    def test_replace_image_tag(self):
        self.config(image_tag_quota=1)
        self.image.tags = ['foo']
        self.assertEqual(1, len(self.image.tags))

    def test_replace_too_many_image_tags(self):
        self.config(image_tag_quota=0)

        exc = self.assertRaises(exception.ImageTagLimitExceeded,
                                setattr, self.image, 'tags', ['foo', 'bar'])
        self.assertIn('Attempted: 2, Maximum: 0',
                      encodeutils.exception_to_unicode(exc))
        self.assertEqual(0, len(self.image.tags))

    def test_replace_unlimited_image_tags(self):
        self.config(image_tag_quota=-1)
        self.image.tags = ['foo']
        self.assertEqual(1, len(self.image.tags))

    def test_add_image_tag(self):
        self.config(image_tag_quota=1)
        self.image.tags.add('foo')
        self.assertEqual(1, len(self.image.tags))

    def test_add_too_many_image_tags(self):
        self.config(image_tag_quota=1)
        self.image.tags.add('foo')
        exc = self.assertRaises(exception.ImageTagLimitExceeded,
                                self.image.tags.add, 'bar')
        self.assertIn('Attempted: 2, Maximum: 1',
                      encodeutils.exception_to_unicode(exc))

    def test_add_unlimited_image_tags(self):
        self.config(image_tag_quota=-1)
        self.image.tags.add('foo')
        self.assertEqual(1, len(self.image.tags))

    def test_remove_image_tag_while_over_quota(self):
        self.config(image_tag_quota=1)
        self.image.tags.add('foo')
        self.assertEqual(1, len(self.image.tags))
        self.config(image_tag_quota=0)
        self.image.tags.remove('foo')
        self.assertEqual(0, len(self.image.tags))


class TestQuotaImageTagsProxy(test_utils.BaseTestCase):

    def setUp(self):
        super(TestQuotaImageTagsProxy, self).setUp()

    def test_add(self):
        proxy = glance.quota.QuotaImageTagsProxy(set([]))
        proxy.add('foo')
        self.assertIn('foo', proxy)

    def test_add_too_many_tags(self):
        self.config(image_tag_quota=0)
        proxy = glance.quota.QuotaImageTagsProxy(set([]))
        exc = self.assertRaises(exception.ImageTagLimitExceeded,
                                proxy.add, 'bar')
        self.assertIn('Attempted: 1, Maximum: 0',
                      encodeutils.exception_to_unicode(exc))

    def test_equals(self):
        proxy = glance.quota.QuotaImageTagsProxy(set([]))
        self.assertEqual(set([]), proxy)

    def test_not_equals(self):
        proxy = glance.quota.QuotaImageTagsProxy(set([]))
        self.assertNotEqual('foo', proxy)

    def test_contains(self):
        proxy = glance.quota.QuotaImageTagsProxy(set(['foo']))
        self.assertIn('foo', proxy)

    def test_len(self):
        proxy = glance.quota.QuotaImageTagsProxy(set(['foo',
                                                      'bar',
                                                      'baz',
                                                      'niz']))
        self.assertEqual(4, len(proxy))

    def test_iter(self):
        items = set(['foo', 'bar', 'baz', 'niz'])
        proxy = glance.quota.QuotaImageTagsProxy(items.copy())
        self.assertEqual(4, len(items))
        for item in proxy:
            items.remove(item)
        self.assertEqual(0, len(items))

    def test_tags_attr_no_loop(self):
        proxy = glance.quota.QuotaImageTagsProxy(None)
        self.assertEqual(set([]), proxy.tags)

    def test_tags_deepcopy(self):
        proxy = glance.quota.QuotaImageTagsProxy(set(['a', 'b']))
        proxy_copy = copy.deepcopy(proxy)
        self.assertEqual(set(['a', 'b']), proxy_copy.tags)
        self.assertIn('a', proxy_copy)
        # remove is a found via __getattr__
        proxy_copy.remove('a')
        self.assertNotIn('a', proxy_copy)

    def test_tags_delete(self):
        proxy = glance.quota.QuotaImageTagsProxy(set(['a', 'b']))
        self.assertEqual(set(['a', 'b']), proxy.tags)
        del proxy.tags
        self.assertIsNone(proxy.tags)


class TestImageMemberQuotas(test_utils.BaseTestCase):

    def setUp(self):
        super(TestImageMemberQuotas, self).setUp()
        db_api = unit_test_utils.FakeDB()
        store_api = unit_test_utils.FakeStoreAPI()
        store = unit_test_utils.FakeStoreUtils(store_api)
        context = FakeContext()
        self.image = mock.Mock()
        self.base_image_member_factory = mock.Mock()
        self.image_member_factory = glance.quota.ImageMemberFactoryProxy(
            self.base_image_member_factory, context,
            db_api, store)

    def test_new_image_member(self):
        self.config(image_member_quota=1)

        self.image_member_factory.new_image_member(self.image,
                                                   'fake_id')
        nim = self.base_image_member_factory.new_image_member
        nim.assert_called_once_with(self.image, 'fake_id')

    def test_new_image_member_unlimited_members(self):
        self.config(image_member_quota=-1)

        self.image_member_factory.new_image_member(self.image,
                                                   'fake_id')
        nim = self.base_image_member_factory.new_image_member
        nim.assert_called_once_with(self.image, 'fake_id')

    def test_new_image_member_too_many_members(self):
        self.config(image_member_quota=0)

        self.assertRaises(exception.ImageMemberLimitExceeded,
                          self.image_member_factory.new_image_member,
                          self.image, 'fake_id')


class TestImageLocationQuotas(test_utils.BaseTestCase):

    def setUp(self):
        super(TestImageLocationQuotas, self).setUp()
        self.base_image = mock.Mock()
        self.base_image.locations = []
        self.base_image.size = 1
        self.base_image.extra_properties = {}
        self.image = glance.quota.ImageProxy(self.base_image,
                                             mock.Mock(),
                                             mock.Mock(),
                                             mock.Mock())

        self.image_repo_mock = mock.Mock()
        self.image_repo_proxy = glance.quota.ImageRepoProxy(
            self.image_repo_mock,
            mock.Mock(),
            mock.Mock(),
            mock.Mock())

    def test_replace_image_location(self):
        self.config(image_location_quota=1)
        self.image.locations = [{"url": "file:///fake.img.tar.gz",
                                 "metadata": {}
                                 }]
        self.assertEqual(1, len(self.image.locations))

    def test_replace_too_many_image_locations(self):
        self.config(image_location_quota=1)
        self.image.locations = [{"url": "file:///fake.img.tar.gz",
                                 "metadata": {}}
                                ]
        locations = [
            {"url": "file:///fake1.img.tar.gz", "metadata": {}},
            {"url": "file:///fake2.img.tar.gz", "metadata": {}},
            {"url": "file:///fake3.img.tar.gz", "metadata": {}}
        ]
        exc = self.assertRaises(exception.ImageLocationLimitExceeded,
                                setattr, self.image, 'locations', locations)
        self.assertIn('Attempted: 3, Maximum: 1',
                      encodeutils.exception_to_unicode(exc))
        self.assertEqual(1, len(self.image.locations))

    def test_replace_unlimited_image_locations(self):
        self.config(image_location_quota=-1)
        self.image.locations = [{"url": "file:///fake.img.tar.gz",
                                 "metadata": {}}
                                ]
        self.assertEqual(1, len(self.image.locations))

    def test_add_image_location(self):
        self.config(image_location_quota=1)
        location = {"url": "file:///fake.img.tar.gz", "metadata": {}}
        self.image.locations.append(location)
        self.assertEqual(1, len(self.image.locations))

    def test_add_too_many_image_locations(self):
        self.config(image_location_quota=1)
        location1 = {"url": "file:///fake1.img.tar.gz", "metadata": {}}
        self.image.locations.append(location1)
        location2 = {"url": "file:///fake2.img.tar.gz", "metadata": {}}
        exc = self.assertRaises(exception.ImageLocationLimitExceeded,
                                self.image.locations.append, location2)
        self.assertIn('Attempted: 2, Maximum: 1',
                      encodeutils.exception_to_unicode(exc))

    def test_add_unlimited_image_locations(self):
        self.config(image_location_quota=-1)
        location1 = {"url": "file:///fake1.img.tar.gz", "metadata": {}}
        self.image.locations.append(location1)
        self.assertEqual(1, len(self.image.locations))

    def test_remove_image_location_while_over_quota(self):
        self.config(image_location_quota=1)
        location1 = {"url": "file:///fake1.img.tar.gz", "metadata": {}}
        self.image.locations.append(location1)
        self.assertEqual(1, len(self.image.locations))
        self.config(image_location_quota=0)
        self.image.locations.remove(location1)
        self.assertEqual(0, len(self.image.locations))


class TestImageKeystoneQuota(test_utils.BaseTestCase):
    def setUp(self):
        super(TestImageKeystoneQuota, self).setUp()

        default_limits = {
            ks_quota.QUOTA_IMAGE_SIZE_TOTAL: 500,
            'another_limit': 2,
        }
        ksqf = glance_fixtures.KeystoneQuotaFixture(**default_limits)

        self.useFixture(ksqf)

        self.db_api = unit_test_utils.FakeDB()
        self.useFixture(fixtures.MockPatch('glance.quota.keystone.db',
                                           self.db_api))

    def _create_fake_image(self, context, size):
        location_count = 2
        locations = []
        for i in range(location_count):
            locations.append({'url': 'file:///g/there/it/is%d' % i,
                              'status': 'active',
                              'metadata': {}})
        image_values = {'id': str(uuid.uuid4()), 'owner': context.owner,
                        'status': 'active', 'size': size * units.Mi,
                        'locations': locations}
        self.db_api.image_create(context, image_values)

    def test_enforce_overquota(self):
        # Check that a single large image with multiple locations will
        # trip the quota check.
        self.config(use_keystone_limits=True)
        context = FakeContext()
        self._create_fake_image(context, 300)
        exc = self.assertRaises(exception.LimitExceeded,
                                ks_quota.enforce_image_size_total,
                                context, context.owner)
        self.assertIn('image_size_total is over limit of 500', str(exc))

    def test_enforce_overquota_with_delta(self):
        # Check that delta is honored, if used.
        self.config(use_keystone_limits=True)
        context = FakeContext()
        self._create_fake_image(context, 200)
        ks_quota.enforce_image_size_total(context, context.owner)
        ks_quota.enforce_image_size_total(context, context.owner,
                                          delta=50)
        self.assertRaises(exception.LimitExceeded,
                          ks_quota.enforce_image_size_total,
                          context, context.owner, delta=200)

    def test_enforce_overquota_disabled(self):
        # Just like the overquota case above, but without being enabled,
        # so no failure
        self.config(use_keystone_limits=False)
        context = FakeContext()
        self._create_fake_image(context, 300)
        # Does not raise because keystone limits are disabled
        ks_quota.enforce_image_size_total(context, context.owner)

    def test_enforce_overquota_multiple(self):
        # Check that multiple images with a combined amount
        # (2*2*150=600) over the quota will trip the quota check.
        self.config(use_keystone_limits=True)
        context = FakeContext()
        self._create_fake_image(context, 150)
        self._create_fake_image(context, 150)
        exc = self.assertRaises(exception.LimitExceeded,
                                ks_quota.enforce_image_size_total,
                                context, context.owner)
        self.assertIn('image_size_total is over limit of 500', str(exc))

    def test_enforce_underquota(self):
        self.config(use_keystone_limits=True)
        context = FakeContext()
        self._create_fake_image(context, 100)
        # We are under quota, so no exception expected
        ks_quota.enforce_image_size_total(context, context.owner)

    def test_enforce_underquota_with_others_over_quota(self):
        self.config(use_keystone_limits=True)
        # Put the first tenant over quota
        context = FakeContext()
        self._create_fake_image(context, 300)
        self._create_fake_image(context, 300)

        # Create an image for another tenant that is not over quota
        other_context = FakeContext()
        other_context.owner = 'someone_else'
        self._create_fake_image(other_context, 100)
        # This tenant should pass the quota check, because it is under quota,
        # even though the other is over.
        ks_quota.enforce_image_size_total(other_context, other_context.owner)

    def test_enforce_multiple_limits_under_quota(self):
        self.config(use_keystone_limits=True)
        context = FakeContext()
        # Make sure that we can call the multi-limit handler and pass when
        # we are under quota.
        ks_quota._enforce_some(context, context.owner,
                               {ks_quota.QUOTA_IMAGE_SIZE_TOTAL: lambda: 200,
                                'another_limit': lambda: 1},
                               {'another_limit': 1})

    def test_enforce_multiple_limits_over_quota(self):
        self.config(use_keystone_limits=True)
        context = FakeContext()
        # Make sure that even if one of a multi-limit call is over
        # quota, we get the exception.
        self.assertRaises(exception.LimitExceeded,
                          ks_quota._enforce_some,
                          context, context.owner,
                          {ks_quota.QUOTA_IMAGE_SIZE_TOTAL: lambda: 200,
                           'another_limit': lambda: 1},
                          {'another_limit': 5})

    @mock.patch('oslo_limit.limit.Enforcer')
    @mock.patch.object(ks_quota, 'LOG')
    def test_oslo_limit_config_fail(self, mock_LOG, mock_enforcer):
        self.config(use_keystone_limits=True)
        mock_enforcer.return_value.enforce.side_effect = (
            ol_exc.SessionInitError('test'))
        context = FakeContext()
        self._create_fake_image(context, 100)
        self.assertRaises(ol_exc.SessionInitError,
                          ks_quota.enforce_image_size_total,
                          context, context.owner)
        mock_LOG.error.assert_called_once_with(
            'Failed to initialize oslo_limit, likely due to '
            'incorrect or insufficient configuration: %(err)s',
            {'err': "Can't initialise OpenStackSDK session: test."})



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Functional test case that tests logging output"""

import http.client as http
import os
import stat

import httplib2

from glance.tests import functional


class TestLogging(functional.FunctionalTest):

    """Functional tests for Glance's logging output"""

    def test_debug(self):
        """
        Test logging output proper when debug is on.
        """
        self.cleanup()
        self.start_servers()

        # The default functional test case has both debug on. Let's verify
        # that debug statements appear in the API logs.

        self.assertTrue(os.path.exists(self.api_server.log_file))

        with open(self.api_server.log_file, 'r') as f:
            api_log_out = f.read()

        self.assertIn('DEBUG glance', api_log_out)

        self.stop_servers()

    def test_no_debug(self):
        """
        Test logging output proper when debug is off.
        """
        self.cleanup()
        self.start_servers(debug=False)

        self.assertTrue(os.path.exists(self.api_server.log_file))

        with open(self.api_server.log_file, 'r') as f:
            api_log_out = f.read()

        self.assertNotIn('DEBUG glance', api_log_out)
        self.stop_servers()

    def assertNotEmptyFile(self, path):
        self.assertTrue(os.path.exists(path))
        self.assertNotEqual(os.stat(path)[stat.ST_SIZE], 0)

    def test_logrotate(self):
        """
        Test that we notice when our log file has been rotated
        """

        # Moving in-use files is not supported on Windows.
        # The log handler itself may be configured to rotate files.
        if os.name == 'nt':
            raise self.skipException("Unsupported platform.")

        self.cleanup()
        self.start_servers()

        self.assertNotEmptyFile(self.api_server.log_file)

        os.rename(self.api_server.log_file, self.api_server.log_file + ".1")

        path = "http://%s:%d/" % ("127.0.0.1", self.api_port)
        response, content = httplib2.Http().request(path, 'GET')
        self.assertEqual(http.MULTIPLE_CHOICES, response.status)

        self.assertNotEmptyFile(self.api_server.log_file)

        self.stop_servers()



# Copyright 2022 OVHCloud
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from unittest import mock

from glance_store import backend
from oslo_config import cfg
from taskflow.types import failure

from glance.async_.flows import api_image_import
import glance.common.exception
from glance import domain
import glance.tests.unit.utils as unit_test_utils
import glance.tests.utils as test_utils

CONF = cfg.CONF


TENANT1 = '6838eb7b-6ded-434a-882c-b344c77fe8df'


class TestBaseDownloadTask(test_utils.BaseTestCase):

    def setUp(self):
        super(TestBaseDownloadTask, self).setUp()

        self.config(node_staging_uri='/tmp/staging')
        self.image_repo = mock.MagicMock()
        self.image_id = mock.MagicMock()
        self.uri = mock.MagicMock()
        self.plugin_name = 'FakeBaseDownload'
        self.task_factory = domain.TaskFactory()

        task_input = {
            "import_req": {
                'method': {
                    'name': 'web_download',
                    'uri': 'http://cloud.foo/image.qcow2'
                }
            }
        }
        task_ttl = CONF.task.task_time_to_live

        self.task_type = 'import'
        request_id = 'fake_request_id'
        user_id = 'fake_user'
        self.task = self.task_factory.new_task(self.task_type, TENANT1,
                                               self.image_id, user_id,
                                               request_id,
                                               task_time_to_live=task_ttl,
                                               task_input=task_input)

        self.task_id = self.task.task_id
        self.action_wrapper = api_image_import.ImportActionWrapper(
            self.image_repo, self.image_id, self.task_id)
        self.image_repo.get.return_value = mock.MagicMock(
            extra_properties={'os_glance_import_task': self.task_id})
        self.base_download_task = unit_test_utils.FakeBaseDownloadPlugin(
            self.task.task_id, self.task_type, self.action_wrapper,
            ['foo'], self.plugin_name)
        self.base_download_task._path = "/path/to_downloaded_data"

    def test_base_download_node_staging_uri_is_none(self):
        self.config(node_staging_uri=None)
        self.assertRaises(glance.common.exception.BadTaskConfiguration,
                          unit_test_utils.FakeBaseDownloadPlugin,
                          self.task.task_id, self.task_type, self.uri,
                          self.action_wrapper, ['foo'])

    @mock.patch.object(cfg.ConfigOpts, "set_override")
    def test_base_download_node_store_initialization_failed(
            self, mock_override):
        with mock.patch.object(backend, '_load_store') as mock_load_store:
            mock_load_store.return_value = None
            self.assertRaises(glance.common.exception.BadTaskConfiguration,
                              unit_test_utils.FakeBaseDownloadPlugin,
                              self.task.task_id, self.task_type, self.uri,
                              self.action_wrapper, ['foo'])
            mock_override.assert_called()

    def test_base_download_delete_staging_image_not_exist(self):
        staging_path = "file:///tmp/staging/temp-image"
        delete_from_fs_task = api_image_import._DeleteFromFS(
            self.task.task_id, self.task_type)
        with mock.patch.object(os.path, "exists") as mock_exists:
            mock_exists.return_value = False
            with mock.patch.object(os, "unlink") as mock_unlik:
                delete_from_fs_task.execute(staging_path)

                self.assertEqual(1, mock_exists.call_count)
                self.assertEqual(0, mock_unlik.call_count)

    @mock.patch.object(os.path, "exists")
    def test_base_download_delete_staging_image_failed(self, mock_exists):
        mock_exists.return_value = True
        staging_path = "file:///tmp/staging/temp-image"
        delete_from_fs_task = api_image_import._DeleteFromFS(
            self.task.task_id, self.task_type)
        with mock.patch.object(os, "unlink") as mock_unlink:
            try:
                delete_from_fs_task.execute(staging_path)
            except OSError:
                self.assertEqual(1, mock_unlink.call_count)

            self.assertEqual(1, mock_exists.call_count)

    @mock.patch.object(os.path, "exists")
    def test_base_download_delete_staging_image_succeed(self, mock_exists):
        mock_exists.return_value = True
        staging_path = "file:///tmp/staging/temp-image"
        delete_from_fs_task = api_image_import._DeleteFromFS(
            self.task.task_id, self.task_type)
        with mock.patch.object(os, "unlink") as mock_unlik:
            delete_from_fs_task.execute(staging_path)
            self.assertEqual(1, mock_exists.call_count)
            self.assertEqual(1, mock_unlik.call_count)

    @mock.patch(
        "glance.async_.flows._internal_plugins.base_download.store_api")
    def test_base_download_revert_with_failure(self, mock_store_api):
        image = self.image_repo.get.return_value
        image.extra_properties['os_glance_importing_to_stores'] = 'foo'
        image.extra_properties['os_glance_failed_import'] = ''

        self.base_download_task.execute = mock.MagicMock(
            side_effect=glance.common.exception.ImportTaskError)
        self.base_download_task.revert(None)
        mock_store_api.delete_from_backend.assert_called_once_with(
            "/path/to_downloaded_data")
        self.assertEqual(1, self.image_repo.save.call_count)
        self.assertEqual(
            '', image.extra_properties['os_glance_importing_to_stores'])
        self.assertEqual(
            'foo', image.extra_properties['os_glance_failed_import'])

    @mock.patch(
        "glance.async_.flows._internal_plugins.base_download.store_api")
    def test_base_download_revert_without_failure_multi_store(self,
                                                              mock_store_api):
        enabled_backends = {
            'fast': 'file',
            'cheap': 'file'
        }
        self.config(enabled_backends=enabled_backends)

        self.base_download_task.revert("/path/to_downloaded_data")
        mock_store_api.delete.assert_called_once_with(
            "/path/to_downloaded_data", None)

    @mock.patch(
        "glance.async_.flows._internal_plugins.base_download.store_api")
    def test_base_download_revert_with_failure_without_path(self,
                                                            mock_store_api):
        image = self.image_repo.get.return_value
        image.status = 'importing'
        image.extra_properties['os_glance_importing_to_stores'] = 'foo'
        image.extra_properties['os_glance_failed_import'] = ''
        result = failure.Failure.from_exception(
            glance.common.exception.ImportTaskError())

        self.base_download_task._path = None
        self.base_download_task.revert(result)
        mock_store_api.delete_from_backend.assert_not_called()

        # NOTE(danms): Since we told revert that we were the problem,
        # we should have updated the image status and moved the stores
        # to the failed list.
        self.image_repo.save.assert_called_once_with(image, 'importing')
        self.assertEqual('queued', image.status)
        self.assertEqual(
            '', image.extra_properties['os_glance_importing_to_stores'])
        self.assertEqual(
            'foo', image.extra_properties['os_glance_failed_import'])

    @mock.patch(
        "glance.async_.flows._internal_plugins.base_download.store_api")
    def test_base_download_revert_with_failure_with_path(self, mock_store_api):
        result = failure.Failure.from_exception(
            glance.common.exception.ImportTaskError())

        self.base_download_task.revert(result)
        mock_store_api.delete_from_backend.assert_called_once_with(
            "/path/to_downloaded_data")

    @mock.patch(
        "glance.async_.flows._internal_plugins.base_download.store_api")
    def test_base_download_delete_fails_on_revert(self, mock_store_api):
        result = failure.Failure.from_exception(
            glance.common.exception.ImportTaskError())
        mock_store_api.delete_from_backend.side_effect = Exception

        # this will verify that revert does not break because of failure
        # while deleting data in staging area
        self.base_download_task.revert(result)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import versionutils
from oslo_policy import policy

from glance.policies import base


DEPRECATED_REASON = """
The metadata API now supports project scope and default roles.
"""


metadef_policies = [
    policy.RuleDefault(name="metadef_default", check_str=""),
    policy.RuleDefault(name="metadef_admin",
                       check_str="rule:context_is_admin"),
    policy.DocumentedRuleDefault(
        name="get_metadef_namespace",
        check_str=base.ADMIN_OR_PROJECT_READER_GET_NAMESPACE,
        scope_types=['project'],
        description="Get a specific namespace.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}',
             'method': 'GET'}
        ],
        deprecated_rule=policy.DeprecatedRule(
            name="get_metadef_namespace", check_str="rule:metadef_default",
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.XENA
        ),
    ),
    policy.DocumentedRuleDefault(
        name="get_metadef_namespaces",
        check_str=base.ADMIN_OR_PROJECT_READER,
        scope_types=['project'],
        description="List namespace.",
        operations=[
            {'path': '/v2/metadefs/namespaces',
             'method': 'GET'}
        ],
        deprecated_rule=policy.DeprecatedRule(
            name="get_metadef_namespaces", check_str="rule:metadef_default",
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.XENA
        ),
    ),
    policy.DocumentedRuleDefault(
        name="modify_metadef_namespace",
        check_str="rule:metadef_admin",
        scope_types=['project'],
        description="Modify an existing namespace.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}',
             'method': 'PUT'}
        ],
    ),
    policy.DocumentedRuleDefault(
        name="add_metadef_namespace",
        check_str="rule:metadef_admin",
        scope_types=['project'],
        description="Create a namespace.",
        operations=[
            {'path': '/v2/metadefs/namespaces',
             'method': 'POST'}
        ],
    ),
    policy.DocumentedRuleDefault(
        name="delete_metadef_namespace",
        check_str="rule:metadef_admin",
        scope_types=['project'],
        description="Delete a namespace.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}',
             'method': 'DELETE'}
        ],
    ),

    policy.DocumentedRuleDefault(
        name="get_metadef_object",
        check_str=base.ADMIN_OR_PROJECT_READER_GET_NAMESPACE,
        scope_types=['project'],
        description="Get a specific object from a namespace.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/objects'
                     '/{object_name}',
             'method': 'GET'}
        ],
        deprecated_rule=policy.DeprecatedRule(
            name="get_metadef_object", check_str="rule:metadef_default",
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.XENA
        ),
    ),
    policy.DocumentedRuleDefault(
        name="get_metadef_objects",
        check_str=base.ADMIN_OR_PROJECT_READER_GET_NAMESPACE,
        scope_types=['project'],
        description="Get objects from a namespace.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/objects',
             'method': 'GET'}
        ],
        deprecated_rule=policy.DeprecatedRule(
            name="get_metadef_objects", check_str="rule:metadef_default",
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.XENA
        ),
    ),
    policy.DocumentedRuleDefault(
        name="modify_metadef_object",
        check_str="rule:metadef_admin",
        scope_types=['project'],
        description="Update an object within a namespace.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/objects'
                     '/{object_name}',
             'method': 'PUT'}
        ],
    ),
    policy.DocumentedRuleDefault(
        name="add_metadef_object",
        check_str="rule:metadef_admin",
        scope_types=['project'],
        description="Create an object within a namespace.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/objects',
             'method': 'POST'}
        ],
    ),
    policy.DocumentedRuleDefault(
        name="delete_metadef_object",
        check_str="rule:metadef_admin",
        scope_types=['project'],
        description="Delete an object within a namespace.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/objects'
                     '/{object_name}',
             'method': 'DELETE'}
        ],
    ),

    policy.DocumentedRuleDefault(
        name="list_metadef_resource_types",
        check_str=base.ADMIN_OR_PROJECT_READER_GET_NAMESPACE,
        scope_types=['project'],
        description="List meta definition resource types.",
        operations=[
            {'path': '/v2/metadefs/resource_types',
             'method': 'GET'}
        ],
        deprecated_rule=policy.DeprecatedRule(
            name="list_metadef_resource_types",
            check_str="rule:metadef_default",
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.XENA
        ),
    ),
    policy.DocumentedRuleDefault(
        name="get_metadef_resource_type",
        check_str=base.ADMIN_OR_PROJECT_READER_GET_NAMESPACE,
        scope_types=['project'],
        description="Get meta definition resource types associations.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/resource_types',
             'method': 'GET'}
        ],
        deprecated_rule=policy.DeprecatedRule(
            name="get_metadef_resource_type",
            check_str="rule:metadef_default",
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.XENA
        ),
    ),
    policy.DocumentedRuleDefault(
        name="add_metadef_resource_type_association",
        check_str="rule:metadef_admin",
        scope_types=['project'],
        description="Create meta definition resource types association.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/resource_types',
             'method': 'POST'}
        ],
    ),
    policy.DocumentedRuleDefault(
        name="remove_metadef_resource_type_association",
        check_str="rule:metadef_admin",
        scope_types=['project'],
        description="Delete meta definition resource types association.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/resource_types'
                     '/{name}',
             'method': 'POST'}
        ],
    ),

    policy.DocumentedRuleDefault(
        name="get_metadef_property",
        check_str=base.ADMIN_OR_PROJECT_READER_GET_NAMESPACE,
        scope_types=['project'],
        description="Get a specific meta definition property.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/properties'
                     '/{property_name}',
             'method': 'GET'}
        ],
        deprecated_rule=policy.DeprecatedRule(
            name="get_metadef_property",
            check_str="rule:metadef_default",
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.XENA
        ),
    ),
    policy.DocumentedRuleDefault(
        name="get_metadef_properties",
        check_str=base.ADMIN_OR_PROJECT_READER_GET_NAMESPACE,
        scope_types=['project'],
        description="List meta definition properties.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/properties',
             'method': 'GET'}
        ],
        deprecated_rule=policy.DeprecatedRule(
            name="get_metadef_properties",
            check_str="rule:metadef_default",
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.XENA
        ),
    ),
    policy.DocumentedRuleDefault(
        name="modify_metadef_property",
        check_str="rule:metadef_admin",
        scope_types=['project'],
        description="Update meta definition property.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/properties'
                     '/{property_name}',
             'method': 'GET'}
        ],
    ),
    policy.DocumentedRuleDefault(
        name="add_metadef_property",
        check_str="rule:metadef_admin",
        scope_types=['project'],
        description="Create meta definition property.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/properties',
             'method': 'POST'}
        ],
    ),
    policy.DocumentedRuleDefault(
        name="remove_metadef_property",
        check_str="rule:metadef_admin",
        scope_types=['project'],
        description="Delete meta definition property.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/properties'
                     '/{property_name}',
             'method': 'DELETE'}
        ],
    ),

    policy.DocumentedRuleDefault(
        name="get_metadef_tag",
        check_str=base.ADMIN_OR_PROJECT_READER_GET_NAMESPACE,
        scope_types=['project'],
        description="Get tag definition.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/tags'
                     '/{tag_name}',
             'method': 'GET'}
        ],
        deprecated_rule=policy.DeprecatedRule(
            name="get_metadef_tag", check_str="rule:metadef_default",
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.XENA
        ),
    ),
    policy.DocumentedRuleDefault(
        name="get_metadef_tags",
        check_str=base.ADMIN_OR_PROJECT_READER_GET_NAMESPACE,
        scope_types=['project'],
        description="List tag definitions.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/tags',
             'method': 'GET'}
        ],
        deprecated_rule=policy.DeprecatedRule(
            name="get_metadef_tags", check_str="rule:metadef_default",
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.XENA
        ),
    ),
    policy.DocumentedRuleDefault(
        name="modify_metadef_tag",
        check_str="rule:metadef_admin",
        scope_types=['project'],
        description="Update tag definition.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/tags'
                     '/{tag_name}',
             'method': 'PUT'}
        ],
    ),
    policy.DocumentedRuleDefault(
        name="add_metadef_tag",
        check_str="rule:metadef_admin",
        scope_types=['project'],
        description="Add tag definition.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/tags'
                     '/{tag_name}',
             'method': 'POST'}
        ],
    ),
    policy.DocumentedRuleDefault(
        name="add_metadef_tags",
        check_str="rule:metadef_admin",
        scope_types=['project'],
        description="Create tag definitions.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/tags',
             'method': 'POST'}
        ],
    ),
    policy.DocumentedRuleDefault(
        name="delete_metadef_tag",
        check_str="rule:metadef_admin",
        scope_types=['project'],
        description="Delete tag definition.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/tags'
                     '/{tag_name}',
             'method': 'DELETE'}
        ],
    ),
    policy.DocumentedRuleDefault(
        name="delete_metadef_tags",
        check_str="rule:metadef_admin",
        scope_types=['project'],
        description="Delete tag definitions.",
        operations=[
            {'path': '/v2/metadefs/namespaces/{namespace_name}/tags',
             'method': 'DELETE'}
        ],
    )
]


def list_rules():
    return metadef_policies



# Copyright 2015 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_concurrency import processutils as putils
from oslo_log import log as logging
from oslo_utils import encodeutils
from oslo_utils import units
from taskflow import task

from glance.common import exception as glance_exception
from glance.i18n import _LW


LOG = logging.getLogger(__name__)

# NOTE(hemanthm): As reported in the bug #1449062, "qemu-img info" calls can
# be exploited to craft DoS attacks by providing malicious input. The process
# limits defined here are protections against such attacks. This essentially
# limits the CPU time and address space used by the process that executes
# "qemu-img info" command to 2 seconds and 1 GB respectively.
QEMU_IMG_PROC_LIMITS = putils.ProcessLimits(cpu_time=2,
                                            address_space=1 * units.Gi)


class OptionalTask(task.Task):

    def __init__(self, *args, **kwargs):
        super(OptionalTask, self).__init__(*args, **kwargs)
        self.execute = self._catch_all(self.execute)

    def _catch_all(self, func):
        # NOTE(flaper87): Read this comment before calling the MI6
        # Here's the thing, there's no nice way to define "optional"
        # tasks. That is, tasks whose failure shouldn't affect the execution
        # of the flow. The only current "sane" way to do this, is by catching
        # everything and logging. This seems harmless from a taskflow
        # perspective but it is not. There are some issues related to this
        # "workaround":
        #
        # - Task's states will shamelessly lie to us saying the task succeeded.
        #
        # - No revert procedure will be triggered, which means optional tasks,
        # for now, mustn't cause any side-effects because they won't be able to
        # clean them up. If these tasks depend on other task that do cause side
        # effects, a task that cleans those side effects most be registered as
        # well. For example, _ImportToFS, _MyDumbTask, _DeleteFromFS.
        #
        # - Ideally, optional tasks shouldn't `provide` new values unless they
        # are part of an optional flow. Due to the decoration of the execute
        # method, these tasks will need to define the provided methods at
        # class level using `default_provides`.
        #
        #
        # The taskflow team is working on improving this and on something that
        # will provide the ability of defining optional tasks. For now, to lie
        # ourselves we must.
        #
        # NOTE(harlowja): The upstream change that is hopefully going to make
        # this easier/built-in is at: https://review.opendev.org/#/c/271116/
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as exc:
                msg = (_LW("An optional task has failed, "
                           "the failure was: %s") %
                       encodeutils.exception_to_unicode(exc))
                LOG.warning(msg)
        return wrapper


def get_glance_endpoint(context, region, interface):
    """Return glance endpoint depending the input tasks

    """
    # We use the current context to retrieve the image
    catalog = context.service_catalog

    for service in catalog:
        if service['type'] == 'image':
            for endpoint in service['endpoints']:
                if endpoint['region'].lower() == region.lower():
                    return endpoint.get('%sURL' % interface)

    raise glance_exception.GlanceEndpointNotFound(region=region,
                                                  interface=interface)



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Routines for configuring Glance
"""

import logging
import os

from oslo_config import cfg
from oslo_middleware import cors
from oslo_policy import opts
from oslo_policy import policy
from paste import deploy

from glance.i18n import _
from glance.version import version_info as version

paste_deploy_opts = [
    cfg.StrOpt('flavor',
               sample_default='keystone',
               help=_("""
Deployment flavor to use in the server application pipeline.

Provide a string value representing the appropriate deployment
flavor used in the server application pipeline. This is typically
the partial name of a pipeline in the paste configuration file with
the service name removed.

For example, if your paste section name in the paste configuration
file is [pipeline:glance-api-keystone], set ``flavor`` to
``keystone``.

Possible values:
    * String value representing a partial pipeline name.

Related Options:
    * config_file

""")),
    cfg.StrOpt('config_file',
               sample_default='glance-api-paste.ini',
               help=_("""
Name of the paste configuration file.

Provide a string value representing the name of the paste
configuration file to use for configuring pipelines for
server application deployments.

NOTES:
    * Provide the name or the path relative to the glance directory
      for the paste configuration file and not the absolute path.
    * The sample paste configuration file shipped with Glance need
      not be edited in most cases as it comes with ready-made
      pipelines for all common deployment flavors.

If no value is specified for this option, the ``paste.ini`` file
with the prefix of the corresponding Glance service's configuration
file name will be searched for in the known configuration
directories. (For example, if this option is missing from or has no
value set in ``glance-api.conf``, the service will look for a file
named ``glance-api-paste.ini``.) If the paste configuration file is
not found, the service will not start.

Possible values:
    * A string value representing the name of the paste configuration
      file.

Related Options:
    * flavor

""")),
]
image_format_opts = [
    cfg.ListOpt('container_formats',
                default=['ami', 'ari', 'aki', 'bare', 'ovf', 'ova', 'docker',
                         'compressed'],
                help=_("Supported values for the 'container_format' "
                       "image attribute"),
                deprecated_opts=[cfg.DeprecatedOpt('container_formats',
                                                   group='DEFAULT')]),
    cfg.ListOpt('disk_formats',
                default=['ami', 'ari', 'aki', 'vhd', 'vhdx', 'vmdk', 'raw',
                         'qcow2', 'vdi', 'iso', 'ploop'],
                help=_("Supported values for the 'disk_format' "
                       "image attribute"),
                deprecated_opts=[cfg.DeprecatedOpt('disk_formats',
                                                   group='DEFAULT')]),
    cfg.ListOpt('vmdk_allowed_types',
                default=['streamOptimized', 'monolithicSparse'],
                help=_("A list of strings describing allowed VMDK "
                       "'create-type' subformats that will be allowed. "
                       "This is recommended to only include "
                       "single-file-with-sparse-header variants to avoid "
                       "potential host file exposure due to processing named "
                       "extents. If this list is empty, then no VDMK image "
                       "types allowed. Note that this is currently only "
                       "checked during image conversion (if enabled), and "
                       "limits the types of VMDK images we will convert "
                       "from.")),
]
task_opts = [
    cfg.IntOpt('task_time_to_live',
               default=48,
               help=_("Time in hours for which a task lives after, either "
                      "succeeding or failing"),
               deprecated_opts=[cfg.DeprecatedOpt('task_time_to_live',
                                                  group='DEFAULT')]),
    cfg.StrOpt('task_executor',
               default='taskflow',
               help=_("""
Task executor to be used to run task scripts.

Provide a string value representing the executor to use for task
executions. By default, ``TaskFlow`` executor is used.

``TaskFlow`` helps make task executions easy, consistent, scalable
and reliable. It also enables creation of lightweight task objects
and/or functions that are combined together into flows in a
declarative manner.

Possible values:
    * taskflow

Related Options:
    * None

""")),
    cfg.StrOpt('work_dir',
               sample_default='/work_dir',
               help=_("""
Absolute path to the work directory to use for asynchronous
task operations.

The directory set here will be used to operate over images -
normally before they are imported in the destination store.

NOTE: When providing a value for ``work_dir``, please make sure
that enough space is provided for concurrent tasks to run
efficiently without running out of space.

A rough estimation can be done by multiplying the number of
``max_workers`` with an average image size (e.g 500MB). The image
size estimation should be done based on the average size in your
deployment. Note that depending on the tasks running you may need
to multiply this number by some factor depending on what the task
does. For example, you may want to double the available size if
image conversion is enabled. All this being said, remember these
are just estimations and you should do them based on the worst
case scenario and be prepared to act in case they were wrong.

Possible values:
    * String value representing the absolute path to the working
      directory

Related Options:
    * None

""")),
]

common_opts = [
    cfg.BoolOpt('allow_additional_image_properties', default=True,
                deprecated_for_removal=True,
                deprecated_since="Ussuri",
                deprecated_reason=_("""
This option is redundant.  Control custom image property usage via the
'image_property_quota' configuration option.  This option is scheduled
to be removed during the Victoria development cycle.
"""),
                help=_("""
Allow users to add additional/custom properties to images.

Glance defines a standard set of properties (in its schema) that
appear on every image. These properties are also known as
``base properties``. In addition to these properties, Glance
allows users to add custom properties to images. These are known
as ``additional properties``.

By default, this configuration option is set to ``True`` and users
are allowed to add additional properties. The number of additional
properties that can be added to an image can be controlled via
``image_property_quota`` configuration option.

Possible values:
    * True
    * False

Related options:
    * image_property_quota

""")),
    cfg.StrOpt('hashing_algorithm',
               default='sha512',
               help=_("""
Secure hashing algorithm used for computing the 'os_hash_value' property.

This option configures the Glance "multihash", which consists of two
image properties: the 'os_hash_algo' and the 'os_hash_value'.  The
'os_hash_algo' will be populated by the value of this configuration
option, and the 'os_hash_value' will be populated by the hexdigest computed
when the algorithm is applied to the uploaded or imported image data.

The value must be a valid secure hash algorithm name recognized by the
python 'hashlib' library.  You can determine what these are by examining
the 'hashlib.algorithms_available' data member of the version of the
library being used in your Glance installation.  For interoperability
purposes, however, we recommend that you use the set of secure hash
names supplied by the 'hashlib.algorithms_guaranteed' data member because
those algorithms are guaranteed to be supported by the 'hashlib' library
on all platforms.  Thus, any image consumer using 'hashlib' locally should
be able to verify the 'os_hash_value' of the image.

The default value of 'sha512' is a performant secure hash algorithm.

If this option is misconfigured, any attempts to store image data will fail.
For that reason, we recommend using the default value.

Possible values:
    * Any secure hash algorithm name recognized by the Python 'hashlib'
      library

Related options:
    * None

""")),
    cfg.IntOpt('image_member_quota', default=128,
               help=_("""
Maximum number of image members per image.

This limits the maximum of users an image can be shared with. Any negative
value is interpreted as unlimited.

Related options:
    * None

""")),
    cfg.IntOpt('image_property_quota', default=128,
               help=_("""
Maximum number of properties allowed on an image.

This enforces an upper limit on the number of additional properties an image
can have. Any negative value is interpreted as unlimited.

NOTE: This won't have any impact if additional properties are disabled. Please
refer to ``allow_additional_image_properties``.

Related options:
    * ``allow_additional_image_properties``

""")),
    cfg.IntOpt('image_tag_quota', default=128,
               help=_("""
Maximum number of tags allowed on an image.

Any negative value is interpreted as unlimited.

Related options:
    * None

""")),
    cfg.IntOpt('image_location_quota', default=10,
               help=_("""
Maximum number of locations allowed on an image.

Any negative value is interpreted as unlimited.

Related options:
    * None

""")),
    cfg.IntOpt('limit_param_default', default=25, min=1,
               help=_("""
The default number of results to return for a request.

Responses to certain API requests, like list images, may return
multiple items. The number of results returned can be explicitly
controlled by specifying the ``limit`` parameter in the API request.
However, if a ``limit`` parameter is not specified, this
configuration value will be used as the default number of results to
be returned for any API request.

NOTES:
    * The value of this configuration option may not be greater than
      the value specified by ``api_limit_max``.
    * Setting this to a very large value may slow down database
      queries and increase response times. Setting this to a
      very low value may result in poor user experience.

Possible values:
    * Any positive integer

Related options:
    * api_limit_max

""")),
    cfg.IntOpt('api_limit_max', default=1000, min=1,
               help=_("""
Maximum number of results that could be returned by a request.

As described in the help text of ``limit_param_default``, some
requests may return multiple results. The number of results to be
returned are governed either by the ``limit`` parameter in the
request or the ``limit_param_default`` configuration option.
The value in either case, can't be greater than the absolute maximum
defined by this configuration option. Anything greater than this
value is trimmed down to the maximum value defined here.

NOTE: Setting this to a very large value may slow down database
      queries and increase response times. Setting this to a
      very low value may result in poor user experience.

Possible values:
    * Any positive integer

Related options:
    * limit_param_default

""")),
    cfg.BoolOpt('show_image_direct_url', default=False,
                help=_("""
Show direct image location when returning an image.

This configuration option indicates whether to show the direct image
location when returning image details to the user. The direct image
location is where the image data is stored in backend storage. This
image location is shown under the image property ``direct_url``.

When multiple image locations exist for an image, the best location
is displayed based on the location strategy indicated by the
configuration option ``location_strategy``.

NOTES:
    * Revealing image locations can present a GRAVE SECURITY RISK as
      image locations can sometimes include credentials. Hence, this
      is set to ``False`` by default. Set this to ``True`` with
      EXTREME CAUTION and ONLY IF you know what you are doing!
    * If an operator wishes to avoid showing any image location(s)
      to the user, then both this option and
      ``show_multiple_locations`` MUST be set to ``False``.

Possible values:
    * True
    * False

Related options:
    * show_multiple_locations
    * location_strategy

""")),
    # NOTE(flaper87): The policy.yaml file should be updated and the location
    # related rules set to admin only once this option is finally removed.
    # NOTE(rosmaita): Unfortunately, this option is used to gate some code
    # paths; if the location related policies are set admin-only, then no
    # normal users can save or retrieve image data.
    cfg.BoolOpt('show_multiple_locations', default=False,
                deprecated_for_removal=True,
                deprecated_reason=_('Use of this option, deprecated since '
                                    'Newton, is a security risk and will be '
                                    'removed once we figure out a way to '
                                    'satisfy those use cases that currently '
                                    'require it.  An earlier announcement '
                                    'that the same functionality can be '
                                    'achieved with greater granularity by '
                                    'using policies is incorrect.  You cannot '
                                    'work around this option via policy '
                                    'configuration at the present time, '
                                    'though that is the direction we believe '
                                    'the fix will take.  Please keep an eye '
                                    'on the Glance release notes to stay up '
                                    'to date on progress in addressing this '
                                    'issue.'),
                deprecated_since='Newton',
                help=_("""
Show all image locations when returning an image.

This configuration option indicates whether to show all the image
locations when returning image details to the user. When multiple
image locations exist for an image, the locations are ordered based
on the location strategy indicated by the configuration opt
``location_strategy``. The image locations are shown under the
image property ``locations``.

NOTES:
    * Revealing image locations can present a GRAVE SECURITY RISK as
      image locations can sometimes include credentials. Hence, this
      is set to ``False`` by default. Set this to ``True`` with
      EXTREME CAUTION and ONLY IF you know what you are doing!
    * See https://wiki.openstack.org/wiki/OSSN/OSSN-0065 for more
      information.
    * If an operator wishes to avoid showing any image location(s)
      to the user, then both this option and
      ``show_image_direct_url`` MUST be set to ``False``.

Possible values:
    * True
    * False

Related options:
    * show_image_direct_url
    * location_strategy

""")),
    cfg.IntOpt('image_size_cap', default=1099511627776, min=1,
               max=9223372036854775808,
               help=_("""
Maximum size of image a user can upload in bytes.

An image upload greater than the size mentioned here would result
in an image creation failure. This configuration option defaults to
1099511627776 bytes (1 TiB).

NOTES:
    * This value should only be increased after careful
      consideration and must be set less than or equal to
      8 EiB (9223372036854775808).
    * This value must be set with careful consideration of the
      backend storage capacity. Setting this to a very low value
      may result in a large number of image failures. And, setting
      this to a very large value may result in faster consumption
      of storage. Hence, this must be set according to the nature of
      images created and storage capacity available.

Possible values:
    * Any positive number less than or equal to 9223372036854775808

""")),
    cfg.StrOpt('user_storage_quota', default='0',
               help=_("""
Maximum amount of image storage per tenant.

This enforces an upper limit on the cumulative storage consumed by all images
of a tenant across all stores. This is a per-tenant limit.

The default unit for this configuration option is Bytes. However, storage
units can be specified using case-sensitive literals ``B``, ``KB``, ``MB``,
``GB`` and ``TB`` representing Bytes, KiloBytes, MegaBytes, GigaBytes and
TeraBytes respectively. Note that there should not be any space between the
value and unit. Value ``0`` signifies no quota enforcement. Negative values
are invalid and result in errors.

This has no effect if ``use_keystone_limits`` is enabled.

Possible values:
    * A string that is a valid concatenation of a non-negative integer
      representing the storage value and an optional string literal
      representing storage units as mentioned above.

Related options:
    * use_keystone_limits

""")),
    cfg.BoolOpt('use_keystone_limits', default=False,
                help=_("""
Utilize per-tenant resource limits registered in Keystone.

Enabling this feature will cause Glance to retrieve limits set in keystone
for resource consumption and enforce them against API users. Before turning
this on, the limits need to be registered in Keystone or all quotas will be
considered to be zero, and thus reject all new resource requests.

These per-tenant resource limits are independent from the static
global ones configured in this config file. If this is enabled, the
relevant static global limits will be ignored.
""")),
    cfg.HostAddressOpt('pydev_worker_debug_host',
                       sample_default='localhost',
                       help=_("""
Host address of the pydev server.

Provide a string value representing the hostname or IP of the
pydev server to use for debugging. The pydev server listens for
debug connections on this address, facilitating remote debugging
in Glance.

Possible values:
    * Valid hostname
    * Valid IP address

Related options:
    * None

""")),
    cfg.PortOpt('pydev_worker_debug_port',
                default=5678,
                help=_("""
Port number that the pydev server will listen on.

Provide a port number to bind the pydev server to. The pydev
process accepts debug connections on this port and facilitates
remote debugging in Glance.

Possible values:
    * A valid port number

Related options:
    * None

""")),
    cfg.StrOpt('metadata_encryption_key',
               secret=True,
               help=_("""
AES key for encrypting store location metadata.

Provide a string value representing the AES cipher to use for
encrypting Glance store metadata.

NOTE: The AES key to use must be set to a random string of length
16, 24 or 32 bytes.

Possible values:
    * String value representing a valid AES key

Related options:
    * None

""")),
    cfg.StrOpt('digest_algorithm',
               default='sha256',
               help=_("""
Digest algorithm to use for digital signature.

Provide a string value representing the digest algorithm to
use for generating digital signatures. By default, ``sha256``
is used.

To get a list of the available algorithms supported by the version
of OpenSSL on your platform, run the command:
``openssl list-message-digest-algorithms``.
Examples are 'sha1', 'sha256', and 'sha512'.

NOTE: ``digest_algorithm`` is not related to Glance's image signing
and verification. It is only used to sign the universally unique
identifier (UUID) as a part of the certificate file and key file
validation.

Possible values:
    * An OpenSSL message digest algorithm identifier

Relation options:
    * None

""")),
    cfg.StrOpt('node_staging_uri',
               default='file:///tmp/staging/',
               help=_("""
The URL provides location where the temporary data will be stored

This option is for Glance internal use only. Glance will save the
image data uploaded by the user to 'staging' endpoint during the
image import process.

This option does not change the 'staging' API endpoint by any means.

NOTE: It is discouraged to use same path as [task]/work_dir

NOTE: 'file://<absolute-directory-path>' is the only option
api_image_import flow will support for now.

NOTE: The staging path must be on shared filesystem available to all
Glance API nodes.

Possible values:
    * String starting with 'file://' followed by absolute FS path

Related options:
    * [task]/work_dir

""")),
    cfg.ListOpt('enabled_import_methods',
                item_type=cfg.types.String(quotes=True),
                bounds=True,
                default=['glance-direct', 'web-download',
                         'copy-image'],
                help=_("""
    List of enabled Image Import Methods

    'glance-direct', 'copy-image' and 'web-download' are enabled by default.
    'glance-download' is available, but requires federated deployments.

    Related options:
        * [DEFAULT]/node_staging_uri""")),
    cfg.StrOpt('worker_self_reference_url',
               default=None,
               help=_("""
The URL to this worker.

If this is set, other glance workers will know how to contact this one
directly if needed. For image import, a single worker stages the image
and other workers need to be able to proxy the import request to the
right one.

If unset, this will be considered to be `public_endpoint`, which
normally would be set to the same value on all workers, effectively
disabling the proxying behavior.

Possible values:
    * A URL by which this worker is reachable from other workers

Related options:
    * public_endpoint

""")),
]

wsgi_opts = [
    cfg.IntOpt('task_pool_threads',
               default=16,
               min=1,
               help=_("""
The number of threads (per worker process) in the pool for processing
asynchronous tasks. This controls how many asynchronous tasks (i.e. for
image interoperable import) each worker can run at a time. If this is
too large, you *may* have increased memory footprint per worker and/or you
may overwhelm other system resources such as disk or outbound network
bandwidth. If this is too small, image import requests will have to wait
until a thread becomes available to begin processing.""")),
    cfg.StrOpt('python_interpreter',
               default=None,
               help=_("""
Path to the python interpreter to use when spawning external
processes. If left unspecified, this will be sys.executable, which should
be the same interpreter running Glance itself. However, in some situations
(for example, uwsgi) sys.executable may not actually point to a python
interpreter and an alternative value must be set.""")),
]


CONF = cfg.CONF
CONF.register_opts(paste_deploy_opts, group='paste_deploy')
CONF.register_opts(image_format_opts, group='image_format')
CONF.register_opts(task_opts, group='task')
CONF.register_opts(common_opts)
CONF.register_opts(wsgi_opts, group='wsgi')
policy.Enforcer(CONF)


def parse_args(args=None, usage=None, default_config_files=None):
    CONF(args=args,
         project='glance',
         version=version.cached_version_string(),
         usage=usage,
         default_config_files=default_config_files)


def parse_cache_args(args=None):
    config_files = cfg.find_config_files(project='glance', prog='glance-cache')
    parse_args(args=args, default_config_files=config_files)


def _get_deployment_flavor(flavor=None):
    """
    Retrieve the paste_deploy.flavor config item, formatted appropriately
    for appending to the application name.

    :param flavor: if specified, use this setting rather than the
                   paste_deploy.flavor configuration setting
    """
    if not flavor:
        flavor = CONF.paste_deploy.flavor
    return '' if not flavor else ('-' + flavor)


def _get_paste_config_path():
    paste_suffix = '-paste.ini'
    conf_suffix = '.conf'
    if CONF.config_file:
        # Assume paste config is in a paste.ini file corresponding
        # to the last config file
        path = CONF.config_file[-1].replace(conf_suffix, paste_suffix)
    else:
        path = CONF.prog + paste_suffix
    return CONF.find_file(os.path.basename(path))


def _get_deployment_config_file():
    """
    Retrieve the deployment_config_file config item, formatted as an
    absolute pathname.
    """
    path = CONF.paste_deploy.config_file
    if not path:
        path = _get_paste_config_path()
    if not path or not (os.path.isfile(os.path.abspath(path))):
        msg = _("Unable to locate paste config file for %s.") % CONF.prog
        raise RuntimeError(msg)
    return os.path.abspath(path)


def load_paste_app(app_name, flavor=None, conf_file=None):
    """
    Builds and returns a WSGI app from a paste config file.

    We assume the last config file specified in the supplied ConfigOpts
    object is the paste config file, if conf_file is None.

    :param app_name: name of the application to load
    :param flavor: name of the variant of the application to load
    :param conf_file: path to the paste config file

    :raises RuntimeError: when config file cannot be located or application
            cannot be loaded from config file
    """
    # append the deployment flavor to the application name,
    # in order to identify the appropriate paste pipeline
    app_name += _get_deployment_flavor(flavor)

    if not conf_file:
        conf_file = _get_deployment_config_file()

    try:
        logger = logging.getLogger(__name__)
        logger.debug("Loading %(app_name)s from %(conf_file)s",
                     {'conf_file': conf_file, 'app_name': app_name})

        app = deploy.loadapp("config:%s" % conf_file, name=app_name)

        # Log the options used when starting if we're in debug mode...
        if CONF.debug:
            CONF.log_opt_values(logger, logging.DEBUG)

        return app
    except (LookupError, ImportError) as e:
        msg = (_("Unable to load %(app_name)s from "
                 "configuration file %(conf_file)s."
                 "\nGot: %(e)r") % {'app_name': app_name,
                                    'conf_file': conf_file,
                                    'e': e})
        logger.error(msg)
        raise RuntimeError(msg)


def set_config_defaults():
    """This method updates all configuration default values."""
    set_cors_middleware_defaults()

    # TODO(gmann): Remove setting the default value of config policy_file
    # once oslo_policy change the default value to 'policy.yaml'.
    # https://github.com/openstack/oslo.policy/blob/a626ad12fe5a3abd49d70e3e5b95589d279ab578/oslo_policy/opts.py#L49
    DEFAULT_POLICY_FILE = 'policy.yaml'
    opts.set_defaults(cfg.CONF, DEFAULT_POLICY_FILE)


def set_cors_middleware_defaults():
    """Update default configuration options for oslo.middleware."""
    cors.set_defaults(
        allow_headers=['Content-MD5',
                       'X-Image-Meta-Checksum',
                       'X-Storage-Token',
                       'Accept-Encoding',
                       'X-Auth-Token',
                       'X-Identity-Status',
                       'X-Roles',
                       'X-Service-Catalog',
                       'X-User-Id',
                       'X-Tenant-Id',
                       'X-OpenStack-Request-ID'],
        expose_headers=['X-Image-Meta-Checksum',
                        'X-Auth-Token',
                        'X-Subject-Token',
                        'X-Service-Token',
                        'X-OpenStack-Request-ID'],
        allow_methods=['GET',
                       'PUT',
                       'POST',
                       'DELETE',
                       'PATCH']
    )



# Copyright 2015 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

import glance_store
from oslo_concurrency import processutils
from oslo_config import cfg

from glance.async_.flows import introspect
from glance.async_ import utils as async_utils
from glance import domain
import glance.tests.utils as test_utils

CONF = cfg.CONF

UUID1 = 'c80a1a6c-bd1f-41c5-90ee-81afedb1d58d'
TENANT1 = '6838eb7b-6ded-434a-882c-b344c77fe8df'


class TestImportTask(test_utils.BaseTestCase):

    def setUp(self):
        super(TestImportTask, self).setUp()
        self.task_factory = domain.TaskFactory()
        task_input = {
            "import_from": "http://cloud.foo/image.qcow2",
            "import_from_format": "qcow2",
            "image_properties": mock.sentinel.image_properties
        }
        task_ttl = CONF.task.task_time_to_live

        self.task_type = 'import'
        image_id = 'fake_image_id'
        user_id = 'fake_user'
        request_id = 'fake_request_id'
        self.task = self.task_factory.new_task(self.task_type, TENANT1,
                                               image_id, user_id, request_id,
                                               task_time_to_live=task_ttl,
                                               task_input=task_input)

        self.context = mock.Mock()
        self.img_repo = mock.Mock()
        self.task_repo = mock.Mock()
        self.img_factory = mock.Mock()

        glance_store.register_opts(CONF)
        self.config(default_store='file',
                    stores=['file', 'http'],
                    filesystem_store_datadir=self.test_dir,
                    group="glance_store")
        glance_store.create_stores(CONF)

    def test_introspect_success(self):
        image_create = introspect._Introspect(self.task.task_id,
                                              self.task_type,
                                              self.img_repo)

        self.task_repo.get.return_value = self.task
        image_id = mock.sentinel.image_id
        image = mock.MagicMock(image_id=image_id)
        self.img_repo.get.return_value = image

        with mock.patch.object(processutils, 'execute') as exc_mock:
            result = json.dumps({
                "virtual-size": 10737418240,
                "filename": "/tmp/image.qcow2",
                "cluster-size": 65536,
                "format": "qcow2",
                "actual-size": 373030912,
                "format-specific": {
                    "type": "qcow2",
                    "data": {
                        "compat": "0.10"
                    }
                },
                "dirty-flag": False
            })

            exc_mock.return_value = (result, None)
            image_create.execute(image, '/test/path.qcow2')
            self.assertEqual(10737418240, image.virtual_size)

            # NOTE(hemanthm): Assert that process limits are being applied on
            # "qemu-img info" calls. See bug #1449062 for more details.
            kw_args = exc_mock.call_args[1]
            self.assertIn('prlimit', kw_args)
            self.assertEqual(async_utils.QEMU_IMG_PROC_LIMITS,
                             kw_args.get('prlimit'))

    def test_introspect_no_image(self):
        image_create = introspect._Introspect(self.task.task_id,
                                              self.task_type,
                                              self.img_repo)

        self.task_repo.get.return_value = self.task
        image_id = mock.sentinel.image_id
        image = mock.MagicMock(image_id=image_id, virtual_size=None)
        self.img_repo.get.return_value = image

        # NOTE(flaper87): Don't mock, test the error.
        with mock.patch.object(processutils, 'execute') as exc_mock:
            exc_mock.return_value = (None, "some error")
            # NOTE(flaper87): Pls, read the `OptionalTask._catch_all`
            # docs to know why this is commented.
            # self.assertRaises(RuntimeError,
            #                  image_create.execute,
            #                  image, '/test/path.qcow2')
            image_create.execute(image, '/test/path.qcow2')
            self.assertIsNone(image.virtual_size)
